(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory PolyOrder imports RedPair UsableRules Polynomial
begin
 
locale poly_order_carrier = poly_carrier + 
  fixes one :: "'a" ("1")
   and  max0 :: "'a \<Rightarrow> 'a"
   and  description :: "string"
  assumes SN: "SN {(x,y) . wf_elem x \<and> wf_elem y \<and> geq y  zero \<and> gt x y}"
    and max0_id_pos: "\<lbrakk>geq x  zero; wf_elem x\<rbrakk> \<Longrightarrow> max0 x = x"
    and max0_pos: "\<lbrakk>wf_elem x\<rbrakk> \<Longrightarrow> geq (max0 x) zero"
    and max0_mono: "\<lbrakk>x \<ge> y; wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> geq (max0 x) (max0 y)"
    and max0_x: "\<lbrakk>wf_elem x\<rbrakk> \<Longrightarrow> geq (max0 x) x"
    and wf_max0: "wf_elem x \<Longrightarrow> wf_elem (max0 x)"
    and wf_one[simp]: "wf_elem 1"
    and left_one[simp]: "wf_elem x \<Longrightarrow> 1 * x = x"
    and right_one[simp]: "wf_elem x \<Longrightarrow> x * 1 = x"
    and one_geq_zero[simp]: "one \<ge> zero"
    and one_gt_zero[simp]: "one > zero"

text {* now we head for polynomial interpretations *}

types ('f,'a)lpoly_inter = "'f \<Rightarrow> nat \<Rightarrow> ('a \<times> 'a list)"

function eval_termI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a) \<Rightarrow> ('f,'a)lpoly_inter \<Rightarrow> ('v,'a)p_ass \<Rightarrow> ('f,'v)term \<Rightarrow> 'a" 
where "eval_termI _ _ _ _ pi \<alpha> (Var x) = \<alpha> x" 
   |  "eval_termI ze pl ti ma pi \<alpha> (Fun f ts) = (let (a,as) = pi f (length ts) in ma (pl a  (list_sumI ze pl (map (\<lambda> at. ti (fst at) (eval_termI ze pl ti ma pi \<alpha> (snd at))) (zip as ts)))))"
by (pat_completeness, auto)

termination eval_termI
proof
  fix ze pl ti ma pi \<alpha> f and ts :: "('f,'v) term list" and cl const list ct
  assume "ct \<in> set (zip list ts)"
  hence "ct mem (zip list ts)" by (auto simp: mem_iff)
  hence "snd ct mem ts" proof (induct ts arbitrary: list, simp)
    case (Cons a ts)
    show ?case 
    proof (cases list)
      case Nil with Cons show ?thesis by simp
    next
      case (Cons b ll)
      from prems show ?thesis by (cases "ct = (b,a)", auto)
    qed
  qed
  thus "((ze, pl, ti, ma, pi, \<alpha>, snd ct), (ze, pl, ti, ma, pi, \<alpha>, Fun f ts)) \<in> measures [size \<circ> snd \<circ> snd \<circ> snd \<circ> snd \<circ> snd \<circ> snd]" by (auto simp: mem_iff)
next
  show "wf (measures [size \<circ> snd \<circ> snd \<circ> snd \<circ> snd \<circ> snd \<circ> snd])" by simp
qed

context poly_order_carrier
begin

abbreviation eval_term :: "('f,'a)lpoly_inter \<Rightarrow> ('v,'a)p_ass \<Rightarrow> ('f,'v)term \<Rightarrow> 'a" ("_\<guillemotleft>_,_\<guillemotright>" 50)
where "eval_term \<equiv> eval_termI zero plus times max0"

abbreviation wf_lpoly_inter :: "('f,'a)lpoly_inter \<Rightarrow> bool"
where "wf_lpoly_inter pi \<equiv> \<forall> f n. wf_elem (fst (pi f n)) \<and> (\<forall> a. a \<in> set (snd (pi f n)) \<longrightarrow> wf_elem a \<and> geq a zero)"
end

locale poly_order = poly_order_carrier +
  fixes pi :: "('f :: sshow,'a)lpoly_inter"
  assumes wf_pi: "wf_lpoly_inter pi"
begin

lemma wf_terms[simp]: assumes wf_ass: "wf_ass \<alpha>"
  shows "wf_elem (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
proof (induct t)
  case (Var x)
  from wf_ass have "wf_elem (\<alpha> x)" unfolding wf_ass_def by auto
  thus ?case by auto
next
  case (Fun f ts)
  show ?case 
  proof (cases "pi f (length ts)")
    case (Pair a as)
    have "wf_elem (list_sum (map (\<lambda>at. times (fst at) (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts)))" (is "wf_elem ?ls")
    proof (rule wf_list_sum, rule ballI)
      fix p
      assume "p \<in> set (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts))" (is "p \<in> set (map _ ?zip)")
      from this obtain a' t where p: "p = a' * (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" and ab: "(a',t) \<in> set ?zip" by auto
      from ab have a': "a' \<in> set as" and t: "t \<in> set ts" using set_zip_leftD[where x = a' and y = t] set_zip_rightD[where y = t and x = a'] by auto
      from a' Pair have "a' \<in> set (snd (pi f (length ts)))" by auto
      with Fun wf_pi have "wf_elem a'" by auto
      with Fun t p show "wf_elem p" by auto
    qed
    from wf_pi have "wf_elem (fst (pi f (length ts)))" by auto
    with Pair `wf_elem ?ls` have "wf_elem (a + ?ls)" by auto
    thus ?thesis using Pair by (auto simp: wf_max0 max0_pos)
  qed
qed

lemma pos_term[simp]: assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "geq (pi\<guillemotleft>\<alpha>,t\<guillemotright>) zero"
proof (induct t)
  case (Var x)
  from wf_ass pos_ass have "wf_elem (\<alpha> x) \<and> geq (\<alpha> x) zero" unfolding wf_ass_def pos_ass_def by auto
  thus ?case by auto
next
  case (Fun f ts)
  show ?case 
  proof (cases "pi f (length ts)")
    case (Pair a as)
    have "wf_elem (list_sum (map (\<lambda>at. times (fst at) (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts)))" (is "wf_elem ?ls")
    proof (rule wf_list_sum, rule ballI)
      fix p
      assume "p \<in> set (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts))" (is "p \<in> set (map _ ?zip)")
      from this obtain a' t where p: "p = a' * (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" and ab: "(a',t) \<in> set ?zip" by auto
      from ab have a': "a' \<in> set as" and t: "t \<in> set ts" using set_zip_leftD[where x = a' and y = t] set_zip_rightD[where y = t and x = a'] by auto
      from a' Pair have "a' \<in> set (snd (pi f (length ts)))" by auto
      with Fun wf_pi have "wf_elem a'" by auto
      with Fun t p wf_ass wf_pi show "wf_elem p" by auto
    qed
    from wf_pi have "wf_elem (fst (pi f (length ts)))" by auto
    with Pair `wf_elem ?ls` have "wf_elem (a + ?ls)" by auto
    thus ?thesis using Pair by (auto simp: wf_max0 max0_pos)
  qed
qed

definition inter_s :: "('f,'v)trel"
  where "inter_s \<equiv> {(s,t). (\<forall> \<alpha> . wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> gt (pi\<guillemotleft>\<alpha>,s\<guillemotright>) (pi\<guillemotleft>\<alpha>,t\<guillemotright>))}"

definition inter_ns :: "('f,'v)trel"
  where "inter_ns \<equiv> {(s,t). (\<forall> \<alpha> . wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> geq (pi\<guillemotleft>\<alpha>,s\<guillemotright>) (pi\<guillemotleft>\<alpha>,t\<guillemotright>))}"

lemma inter_sn : 
  shows "SN (inter_s :: ('f,'v)trel)" (is "SN ?P")
proof
  fix t :: "('f,'v) term"
  show "SN ?P t" 
  proof (rule ccontr)
    assume "\<not> SN ?P t"
    hence "\<exists>f . f 0 = t \<and> (\<forall>i. (f i,f(Suc i)) \<in> ?P)" by auto
    from this obtain f where "f (0 :: nat) = t" and steps: "(\<forall> i. (f i, f (Suc i)) \<in> ?P)" by auto
    let ?b = zero_ass
    have "\<forall> i. (pi\<guillemotleft>?b,f i\<guillemotright>) > (pi\<guillemotleft>?b, f (Suc i)\<guillemotright>)"
    proof 
      fix i
      from steps have "(f i, f (Suc i)) \<in> inter_s" .. 
      thus "(pi\<guillemotleft>?b, f i\<guillemotright>) >  (pi\<guillemotleft>?b, f (Suc i)\<guillemotright>)"  unfolding inter_s_def 
	by auto
    qed
    from this obtain g where nSN: "\<forall> i. gt ((g i) :: 'a) (g (Suc i)) \<and> wf_elem (g i) \<and> wf_elem (g (Suc i)) \<and> geq (g (Suc i)) zero" by force
    with SN show False by (auto simp: SN_ars_def SN_elt_def)
  qed
qed


lemma inter_compat: "inter_s O inter_ns \<subseteq> inter_s"
proof
  fix st
  assume "st \<in> inter_s O inter_ns"
  thus "st \<in> inter_s" unfolding inter_s_def inter_ns_def
  proof (cases st)
    case (Pair s t)
    assume "st \<in> {(s, t). \<forall>\<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>) } O
                {(s, t). \<forall>\<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge>  pi\<guillemotleft>\<alpha>,t\<guillemotright>) }" (is "_ \<in> ?S O ?NS")
    with Pair obtain u where cond: "(s,u) \<in> ?NS \<and> (u,t) \<in> ?S" by auto
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)" 
    proof (intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)" (is "?pr \<longrightarrow> ?go")
      proof
	assume ?pr
	with cond have both: "(pi\<guillemotleft>\<alpha>,u\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>) \<and> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,u\<guillemotright>)" by auto
	show "(pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)" by (rule compat[where ?t="pi\<guillemotleft>\<alpha>,u\<guillemotright>"], auto simp: both `?pr`)
      qed
    qed
    with Pair show "st \<in> ?S" by simp
  qed
qed

lemma interpr_sub[simp] : assumes wf_ass: "wf_ass \<alpha>"
  shows "(pi\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>) = (pi\<guillemotleft>(\<lambda> x. pi\<guillemotleft>\<alpha>, getSub \<sigma> x\<guillemotright>),t\<guillemotright>)" 
proof (cases \<sigma>)
  case (Sub sig)
  have "(pi\<guillemotleft>\<alpha>, t \<cdot> Sub sig\<guillemotright>) = (pi\<guillemotleft>(\<lambda> x. pi\<guillemotleft>\<alpha>,sig x\<guillemotright>),t\<guillemotright>)" (is "_ = (pi\<guillemotleft>?B, t\<guillemotright>)")
  proof (induct t)
    case (Var x)
    thus ?case by simp
  next
    case (Fun f ss)
    let ?ts = "map (\<lambda> s. s \<cdot> Sub sig) ss"
    have wf_assB: "wf_ass ?B" unfolding wf_ass_def using wf_ass wf_pi by auto
    have map: "(map (eval_term pi \<alpha>) (?ts))
      = (map (eval_term pi ?B) ss)" using Fun by (induct ss, auto)
    let ?i = "pi f (length ss)"
    show ?case 
    proof (cases "?i")
      case (Pair c cs)
      from wf_pi have "\<forall> d \<in> set (snd ?i). wf_elem d" by auto
      with Pair have wf_i: "\<forall> d \<in> set cs. wf_elem d" by auto
      from wf_i map have "map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip cs (map (\<lambda>t. t \<cdot> Sub sig) ss)) = 
	map (\<lambda>at. fst at * (pi\<guillemotleft>\<lambda>x. pi\<guillemotleft>\<alpha>,sig x\<guillemotright>,snd at\<guillemotright>)) (zip cs ss)" 
      proof (induct cs arbitrary: ss, simp)
	case (Cons c cs)
	thus ?case 
	proof (cases ss, simp)
	  case (Cons s sss)
	  from prems show ?thesis by auto
	qed
      qed
      with Pair show ?thesis by auto
    qed
  qed
  with Sub show ?thesis by simp
qed

lemma inter_s_sub: "sub_closed inter_s" 
proof (unfold sub_closed_def inter_s_def, rule subsetI)
  fix st_sig :: "('f,'v)term \<times> ('f,'v)term"
  assume st_sig: "st_sig \<in> CS {(s,t). \<forall>\<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>) }" (is "_ \<in> CS ?P")
  thus "st_sig \<in> ?P"
  proof (cases st_sig)
    case (Pair s_sig t_sig)
    with st_sig have "(s_sig, t_sig) \<in> CS ?P" by simp
    hence "(s_sig, t_sig) \<in> ?P" 
    proof induct
      case (csI s t \<sigma>)
      show ?case 
      proof (simp del:interpr_sub, intro allI)
	fix \<alpha> 
	show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>, s \<cdot> \<sigma>\<guillemotright> > pi\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>)" (is "?wf \<and> ?ps \<longrightarrow> ?go") 
	proof -
	  {
	    assume ?wf and ?ps
	    have wf_assB: "wf_ass (\<lambda>x. pi\<guillemotleft>\<alpha>,getSub \<sigma> x\<guillemotright>)" (is "wf_ass ?bet") unfolding wf_ass_def using `?wf`  by auto
	    have pos_assB: "pos_ass ?bet" unfolding pos_ass_def using `?ps` `?wf`  by auto
	    with prems wf_assB have "pi\<guillemotleft>?bet,s\<guillemotright> > pi\<guillemotleft>?bet,t\<guillemotright>" by force
	    with `?wf` have "?go" by simp
	  }
	  thus ?thesis by blast
	qed
      qed
    qed
    with Pair show ?thesis by simp
  qed
qed

lemma inter_ns_sub: "sub_closed inter_ns" 
proof (unfold sub_closed_def inter_ns_def, rule subsetI)
  fix st_sig :: "('f,'v)term \<times> ('f,'v)term"
  assume st_sig: "st_sig \<in> CS {(s,t). \<forall>\<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>) }" (is "_ \<in> CS ?P")
  thus "st_sig \<in> ?P"
  proof (cases st_sig)
    case (Pair s_sig t_sig)
    with st_sig have "(s_sig, t_sig) \<in> CS ?P" by simp
    hence "(s_sig, t_sig) \<in> ?P" 
    proof induct
      case (csI s t \<sigma>)
      show ?case 
      proof (simp del:interpr_sub, intro allI)
	fix \<alpha> 
	show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>, s \<cdot> \<sigma>\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>)" (is "?wf \<and> ?ps \<longrightarrow> ?go") 
	proof -
	  {
	    assume ?wf and ?ps
	    have wf_assB: "wf_ass (\<lambda>x. pi\<guillemotleft>\<alpha>,getSub \<sigma> x\<guillemotright>)" (is "wf_ass ?bet") unfolding wf_ass_def using `?wf`  by auto
	    have pos_assB: "pos_ass ?bet" unfolding pos_ass_def using `?ps` `?wf`  by auto
	    with prems wf_assB have "pi\<guillemotleft>?bet,s\<guillemotright> \<ge> pi\<guillemotleft>?bet,t\<guillemotright>" by force
	    with `?wf` have "?go" by simp
	  }
	  thus ?thesis by blast
	qed
      qed
    qed
    with Pair show ?thesis by simp
  qed
qed

lemma inter_ns_cc: 
  shows "one_step_ctxt_closed (inter_ns :: ('f,'v)trel)"
proof (unfold one_step_ctxt_closed_def inter_ns_def, (rule allI)+, rule impI)
  fix f :: 'f
    and bef :: "('f,'v) term list" 
    and s :: "('f,'v) term" 
    and t :: "('f,'v) term" 
    and aft :: "('f,'v) term list" 
  let ?P = "{(s,t). \<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)}"
  assume "(s, t) \<in> ?P"
  hence st: "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>, s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>, t\<guillemotright>)" by (simp)
  let ?inter = "pi f (Suc (length bef + length aft))"
  show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> ?P"
  proof (cases "?inter")
    case (Pair c list)
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>, Fun f (bef @ s # aft)\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,Fun f (bef @ t # aft)\<guillemotright>)" (is "\<forall> \<alpha>. ?ass \<alpha> \<longrightarrow> ?ge \<alpha>")
    proof 
      fix \<alpha>
      show "?ass \<alpha> \<longrightarrow> ?ge \<alpha>"
      proof 
	assume "?ass \<alpha>"
	hence wf_ass: "wf_ass \<alpha>" and pos_ass: "pos_ass \<alpha>" by auto
	with st have sta: "pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>" by auto
	from wf_pi have "wf_elem (fst ?inter) \<and> (\<forall> a. a \<in> set (snd ?inter) \<longrightarrow> wf_elem a \<and> geq a zero)"  by auto
	with Pair have wf_c: "wf_elem c" and wf_list: "\<forall> a \<in> set list. wf_elem a" and pos_list: "\<forall> a \<in> set list. geq a zero" by auto
	let ?expr = "\<lambda> list rest.  list_sum (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip list rest))"
	from wf_list pos_list have main: "(?expr list (bef @ s # aft) \<ge> ?expr list (bef @ t # aft)) \<and> wf_elem (?expr list (bef @ s # aft)) \<and> wf_elem (?expr list (bef @ t # aft))"
	proof (induct list arbitrary:bef, simp)
	  case (Cons a as)
	  hence wf_a: "wf_elem a" and pos_a: "a \<ge> 0" and wf_as: "\<forall> a \<in> set as. wf_elem a" and pos_as: "\<forall> a \<in> set as. a \<ge> 0"  by auto
	  thus ?case 
	  proof (cases bef)
	    case Nil
	    have wf_as_aft: "wf_elem (?expr as aft)" 
	    proof (rule wf_list_sum, auto)
	      fix aas aaft
	      assume "(aas,aaft) \<in> set (zip as aft)"
	      hence "aas \<in> set as" using set_zip_leftD[where x = aas and y = aaft] by auto
	      with wf_as have wfaas: "wf_elem aas" by auto
	      show "wf_elem (aas * (pi\<guillemotleft>\<alpha>,aaft\<guillemotright>))" by (auto simp: wf_ass wfaas)
	    qed 
	    show ?thesis
	      by (rule conjI, simp add: Nil, rule plus_left_mono, rule times_right_mono, rule pos_a, rule sta, 
		auto simp: wf_a pos_a wf_ass, rule wf_as_aft, auto simp: Nil wf_as_aft wf_a wf_ass)
	  next
	    case (Cons u us)
	    with prems wf_as pos_as have wf_rec: "wf_elem (?expr as (us @ s # aft)) \<and> wf_elem (?expr as (us @ t # aft))" by best
	    show ?thesis 
	      by (simp add: Cons, rule conjI, rule plus_right_mono, auto simp: prems wf_as pos_as)
	  qed
	qed
	show "?ge \<alpha>" by (simp add: Pair, rule max0_mono, rule plus_right_mono, auto simp: main wf_c)
      qed
    qed
    thus ?thesis by auto
  qed
qed

lemma inter_redpair: "redpair inter_s inter_ns"
  by (rule redpair.intro, rule inter_sn, rule inter_compat, rule inter_ns_cc, rule inter_ns_sub, rule inter_s_sub)
end

function PleftI :: "'a \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a) \<Rightarrow> ('f,'a)lpoly_inter \<Rightarrow> ('f,'v)term \<Rightarrow> ('v,'a)l_poly" 
where "PleftI ze on pl ti mz pi (Var x) = (LPoly ze [(x,on)] )" 
   |  "PleftI ze on pl ti mz pi (Fun f ts) = (let (c, as) = pi f (length ts) in 
              (case (sum_lpolyI ze pl (LPoly c []) (list_sumI (LPoly ze []) (sum_lpolyI ze pl) (map (\<lambda> at . (mul_lpolyI ze ti (fst at)  (PleftI ze on pl ti mz pi (snd at)))) (zip as ts)))) :: ('v,'a)l_poly of
                  LPoly d [] \<Rightarrow> LPoly (mz d) [] | p \<Rightarrow> p))"
by (pat_completeness, auto)

termination PleftI
proof
  fix ze on pl ti mz pi f and ts :: "('f,'v) term list" and cl const list ct
  assume "ct \<in> set (zip list ts)"
  hence "ct mem (zip list ts)" by (auto simp: mem_iff)
  hence "snd ct mem ts" proof (induct ts arbitrary: list, simp)
    case (Cons a ts)
    show ?case 
    proof (cases list)
      case Nil with Cons show ?thesis by simp
    next
      case (Cons b ll)
      from prems show ?thesis by (cases "ct = (b,a)", auto)
    qed
  qed
  thus "((ze,on,pl,ti,mz,pi, snd ct), (ze,on,pl,ti,mz,pi, Fun f ts)) \<in> measures [size \<circ> snd \<circ> snd \<circ> snd \<circ> snd \<circ> snd \<circ> snd]" by (auto simp: mem_iff)
next
  show "wf (measures [size \<circ> snd \<circ> snd \<circ> snd \<circ> snd \<circ> snd \<circ> snd])" by simp
qed

function  PrightI :: "'a \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a) \<Rightarrow> ('f,'a)lpoly_inter \<Rightarrow> ('f,'v)term \<Rightarrow> ('v,'a)l_poly" 
where "PrightI ze on pl ti mz pi (Var x) = (LPoly ze [(x,on)] )" 
   |  "PrightI ze on pl ti mz pi (Fun f ts) = (let (c, as) = pi f (length ts) in 
              (case (sum_lpolyI ze pl (LPoly c []) (list_sumI (LPoly ze []) (sum_lpolyI ze pl) (map (\<lambda> at . (mul_lpolyI ze ti (fst at)  (PrightI ze on pl ti mz pi (snd at)))) (zip as ts)))) :: ('v,'a)l_poly of
                  LPoly d vp \<Rightarrow> LPoly (mz d) vp))"
by (pat_completeness, auto)

termination PrightI
proof
  fix ze on pl ti mz pi f and ts :: "('f,'v) term list" and cl const list ct
  assume "ct \<in> set (zip list ts)"
  hence "ct mem (zip list ts)" by (auto simp: mem_iff)
  hence "snd ct mem ts" proof (induct ts arbitrary: list, simp)
    case (Cons a ts)
    show ?case 
    proof (cases list)
      case Nil with Cons show ?thesis by simp
    next
      case (Cons b ll)
      from prems show ?thesis by (cases "ct = (b,a)", auto)
    qed
  qed
  thus "((ze,on,pl,ti,mz,pi, snd ct), (ze,on,pl,ti,mz,pi, Fun f ts)) \<in> measures [size \<circ> snd \<circ> snd \<circ> snd \<circ> snd \<circ> snd \<circ> snd]" by (auto simp: mem_iff)
next
  show "wf (measures [size \<circ> snd \<circ> snd \<circ> snd \<circ> snd \<circ> snd \<circ> snd])" by simp
qed

fun check_polo_nsI
  where "check_polo_nsI ze on pl ti mz ge pi (s,t) = (
      let left = PleftI ze on pl ti mz pi s
  in  let right = PrightI ze on pl ti mz pi t
  in  (check_lpoly_nsI ze on ge left right 
     \<leftarrow> (\<lambda> s. (''could not ensure '' +++ sshow s +++ '' >= '' +++ sshow t) # s)))"

fun check_polo_sI
  where "check_polo_sI mode ze on pl ti mz ge gt pi (s,t) = (
      let left = PleftI ze on pl ti mz pi s
  in  let right = PrightI ze on pl ti mz pi t
  in (check_lpoly_sI mode ze on ge gt left right
     \<leftarrow> (\<lambda> s. (''could not ensure '' +++ sshow s +++ '' > '' +++ sshow t) # s)))"

context poly_order
begin

abbreviation zero_lpoly :: "('v,'a)l_poly"
where "zero_lpoly \<equiv> LPoly 0 []"

abbreviation c_lpoly :: "'a \<Rightarrow> ('v,'a)l_poly"
where "c_lpoly c \<equiv> LPoly c []"

abbreviation list_sum_lpoly :: "('v,'a)l_poly list \<Rightarrow> ('v,'a)l_poly"
where "list_sum_lpoly \<equiv> list_sumI zero_lpoly sum_lpoly"

lemma wf_list_sum_lpoly : assumes "\<forall> p \<in> set ps. wf_lpoly p"
  shows "wf_lpoly (list_sum_lpoly ps)"
using assms
proof (induct ps, simp add: wf_pvars_def)
  case (Cons p ps)
  show ?case by (simp only: list_sumI.simps, rule wf_sum_lpoly, auto simp: Cons)
qed

lemma pos_list_sum_lpoly : 
  assumes "\<forall> p \<in> set ps. wf_lpoly p"
  and     "\<forall> p \<in> set ps. pos_coeffs p"
  shows "pos_coeffs (list_sum_lpoly ps)"
using assms
proof (induct ps, simp add: pos_coeffs_def pos_pvars_def)
  case (Cons p ps)
  show ?case by (simp only: list_sumI.simps, rule pos_sum_lpoly, auto simp: Cons, rule wf_list_sum_lpoly, auto simp: Cons)
qed

lemma list_sum_lpoly_sound : assumes wf_ps: "\<forall> p \<in> set ps. wf_lpoly p"
  and wf_ass : "wf_ass \<alpha>"
  shows "eval_lpoly \<alpha> (list_sum_lpoly ps) = list_sum (map (eval_lpoly \<alpha>) ps)"
using wf_ps
proof (induct ps, simp)
  case (Cons p ps)
  have "eval_lpoly \<alpha> (list_sum_lpoly (p # ps)) = eval_lpoly \<alpha> (sum_lpoly p (list_sum_lpoly ps))" by auto
  also have "\<dots> = eval_lpoly \<alpha> p + eval_lpoly \<alpha> (list_sum_lpoly ps)" 
    by (rule sum_poly_sound, rule wf_ass, simp add: Cons, rule wf_list_sum_lpoly, auto simp: Cons)
  also have "\<dots> = eval_lpoly \<alpha> p + list_sum (map (eval_lpoly \<alpha>) ps)" using Cons by auto
  finally show ?case by (simp only: list_sumI.simps, auto simp: Cons wf_ass)
qed

abbreviation Pleft :: "('f,'v)term \<Rightarrow> ('v,'a)l_poly"
where "Pleft \<equiv> PleftI zero one plus times max0 pi"

abbreviation Pright :: "('f,'v)term \<Rightarrow> ('v,'a)l_poly"
where "Pright \<equiv> PrightI zero one plus times max0 pi"

lemma Pleft_both :  assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "((pi\<guillemotleft>\<alpha>,t\<guillemotright>) \<ge> eval_lpoly \<alpha> (Pleft t)) \<and> wf_lpoly (Pleft t)" 
proof (induct t)
  case (Var x)
  show ?case using pos_ass wf_ass unfolding pos_ass_def wf_ass_def by (auto simp: wf_pvars_def)
  case (Fun f ts)
  let ?I = "\<lambda> s. pi\<guillemotleft>\<alpha>,s\<guillemotright>"
  let ?E = "\<lambda> p. eval_lpoly \<alpha> p"
  show ?case 
  proof (cases "pi f (length ts)")
    case (Pair c cs)
    from wf_pi have "wf_elem (fst (pi f (length ts))) \<and> (\<forall> c \<in> set (snd (pi f (length ts))). wf_elem c \<and> geq c zero)" by auto
    with Pair have wf_c: "wf_elem c" and wf_cs: "\<forall> c \<in> set cs. wf_elem c" and pos_cs: "\<forall> c \<in> set cs. c \<ge> 0" by auto
    let ?plist = "map (\<lambda> at. mul_lpoly (fst at) (Pleft (snd at))) (zip cs ts)"
    have wf_ps: "\<forall> p \<in> set ?plist. wf_lpoly p"
    proof
      fix p
      assume "p \<in> set ?plist"
      from this obtain a t where p: "p = mul_lpoly a (Pleft t)" and at: "(a,t) \<in> set (zip cs ts)" by auto 
      from at have a: "a \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = a and y = t] set_zip_rightD[where x = a and y = t] by auto
      show "wf_lpoly p" by (simp only: p, rule wf_mul_lpoly, auto simp: wf_cs a t Fun)
    qed
    have wf_sum: "wf_lpoly (list_sum_lpoly ?plist)"
      by (rule wf_list_sum_lpoly, rule wf_ps)
    have "(list_sum (map (\<lambda> at. fst at * (?I (snd at))) (zip cs ts))) \<ge> list_sum (map ?E ?plist)" (is "geq (list_sum ?vlist) _")
    proof (simp only: o_def map_compose[symmetric], rule list_sum_mono, rule ballI)
      fix ct
      assume mem: "ct \<in> set (zip cs ts)"
      from this obtain c t where ct: "ct = (c,t)" by force
      from ct mem have c: "c \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = c and y = t] set_zip_rightD[where x = c and y = t] by auto
      have "((c * (?I t)) \<ge> ?E (mul_lpoly c (Pleft t))) \<and> wf_elem (c * ?I t) \<and> wf_elem (?E (mul_lpoly c (Pleft t)))" 
      proof (rule conjI)
	have id: "?E (mul_lpoly c (Pleft t)) = c * (?E (Pleft t))" by (rule mul_poly_sound, auto simp: c wf_cs Fun t wf_ass)
	show "c * (?I t) \<ge> ?E (mul_lpoly c (Pleft t))" 
	  by (simp only: id, rule times_right_mono, auto simp: c wf_cs Fun  pos_cs t Fun wf_ass, rule wf_eval_lpoly, auto simp: wf_ass t Fun)
      next
	show "wf_elem (c * (?I t)) \<and> wf_elem (?E (mul_lpoly c (Pleft t)))"
	by (rule conjI, rule wf_times, auto simp: c wf_cs wf_ass, rule wf_eval_lpoly, auto simp: wf_ass t Fun, rule wf_mul_lpoly, auto simp: c wf_cs t Fun)
      qed
      with ct show " (fst ct * (pi\<guillemotleft>\<alpha>,snd ct\<guillemotright>) \<ge> eval_lpoly \<alpha> (mul_lpoly (fst ct) (Pleft (snd ct)))) \<and> wf_elem (fst ct * (pi\<guillemotleft>\<alpha>,snd ct\<guillemotright>)) \<and> wf_elem (eval_lpoly \<alpha> (mul_lpoly (fst ct) (Pleft (snd ct))))" by simp
    qed
    also have  "\<dots> = ?E (list_sum_lpoly ?plist)" by (rule list_sum_lpoly_sound[symmetric], rule wf_ps, rule wf_ass)
    finally have part1: "list_sum ?vlist \<ge> ?E (list_sum_lpoly ?plist)" . 

    have wf_csump: "wf_lpoly (sum_lpoly (c_lpoly c) (list_sum_lpoly ?plist))" (is "wf_lpoly ?csump")
      by (rule wf_sum_lpoly, auto simp: wf_c wf_pvars_def, rule wf_sum)
    from Pair Fun have id: "Pleft (Fun f ts) = (case ?csump of LPoly d [] \<Rightarrow> LPoly (max0 d) [] | p \<Rightarrow> p)" by auto
    
    from wf_csump have wf_final: "wf_lpoly (Pleft (Fun f ts))" by (simp only: id, cases ?csump, cases "get_nc_lpoly ?csump", auto simp: wf_pvars_def wf_max0)

    have "?E ?csump = ?E (c_lpoly c) + ?E (list_sum_lpoly ?plist)" by (rule sum_poly_sound, auto simp: wf_ass wf_c wf_pvars_def wf_sum)
    also have "\<dots> = c + ?E (list_sum_lpoly ?plist)" by (simp add: wf_c)
    finally have idr: "?E ?csump = c + ?E (list_sum_lpoly ?plist)" .
    
    have wf_suml: "wf_elem (list_sum ?vlist)"
    proof (rule wf_list_sum, rule ballI)
      fix v
      assume "v \<in> set ?vlist"
      from this obtain c t where v: "v = c * (?I t)" and ct: "(c,t) \<in> set (zip cs ts)" by auto 
      from ct have c: "c \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = c and y = t] set_zip_rightD[where x = c and y = t] by auto
      show "wf_elem v" by (simp only: v, rule wf_times, auto simp: wf_cs c wf_ass)
    qed

    have ge_csum: "c + list_sum ?vlist \<ge> ?E ?csump" 
      by (simp only: idr, rule plus_right_mono, rule part1, rule wf_c, rule wf_suml,
      rule wf_eval_lpoly, rule wf_ass, rule wf_sum)

    have wf_csuml: "wf_elem (c + list_sum ?vlist)" using wf_suml wf_c by auto

    from Pair have left: "?I (Fun f ts) = max0 ( c + list_sum ?vlist)" by auto 
    have "\<dots> \<ge> max0 ( ?E ?csump)"
      by (rule max0_mono, rule ge_csum, rule wf_csuml, rule wf_eval_lpoly, rule wf_ass, rule wf_csump)
    hence part2: "?I (Fun f ts) \<ge> max0 ( ?E ?csump)" by (simp only: left)
    
    have part3: "max0 ( ?E ?csump) \<ge> ?E (Pleft (Fun f ts))"
      using wf_csump by (simp only: id, cases ?csump, cases "get_nc_lpoly ?csump", auto simp: max0_x wf_ass wf_max0)

    have ge_final: "?I (Fun f ts) \<ge> ?E (Pleft (Fun f ts))"
      by (rule geq_trans, rule part2, rule part3, rule wf_terms, rule wf_ass, rule wf_max0, 
	rule wf_eval_lpoly, rule wf_ass, rule wf_csump, rule wf_eval_lpoly, rule wf_ass, rule wf_final)
    
    from wf_final ge_final show ?thesis by auto
  qed
qed
  

lemma Pleft_sound :   assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "(pi\<guillemotleft>\<alpha>,t\<guillemotright>) \<ge> eval_lpoly \<alpha> (Pleft t)"
  using Pleft_both assms by auto

lemma wf_Pleft[simp] : "wf_lpoly (Pleft t)"
  using Pleft_both[where \<alpha> = zero_ass] by auto

lemma pos_eval_pvars: assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  and wf_vas: "wf_pvars vas"
  and pos_vas: "pos_pvars vas"
  shows "eval_pvars \<alpha> vas \<ge> 0"
using wf_vas pos_vas 
proof (induct vas, simp)
  case (Cons xa vas)
  show ?case 
  proof (cases xa)
    case (Pair x a)
    from pos_ass wf_ass have pos_x: "\<alpha> x \<ge> 0" and wf_x: "wf_elem (\<alpha> x)" unfolding pos_ass_def wf_ass_def by auto
    from Pair Cons have pos_a: "a \<ge> 0" and wf_a: "wf_elem a" unfolding wf_pvars_def pos_pvars_def by auto
    have "a * \<alpha> x \<ge> 0 * 0" by (rule geq_trans[where y = "a * 0"], rule times_right_mono, auto simp: wf_a pos_a wf_x pos_x)
    hence pos_ax: "a * \<alpha> x \<ge> 0" by auto
    have wf_ax: "wf_elem (a * \<alpha> x)" by (auto simp: wf_x wf_a)
    from Cons have pos_rec: "eval_pvars \<alpha> vas \<ge> 0" unfolding wf_pvars_def pos_pvars_def by auto
    from Cons have "wf_pvars vas" unfolding wf_pvars_def by auto
    with Cons have wf_rec: "wf_elem (eval_pvars \<alpha> vas)" using wf_ass by auto
    have "a * \<alpha> x + eval_pvars \<alpha> vas \<ge> 0 + 0" by (rule plus_left_right_mono, auto simp: pos_ax wf_ax pos_rec wf_rec)
    thus ?thesis by (simp add: Pair)
  qed
qed
    

lemma Pright_both :  assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "(eval_lpoly \<alpha> (Pright t) \<ge> (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) \<and> wf_lpoly (Pright t) \<and> pos_coeffs (Pright t)" 
proof (induct t)
  case (Var x)
  show ?case using pos_ass wf_ass unfolding pos_ass_def wf_ass_def pos_coeffs_def pos_pvars_def by (auto simp: wf_pvars_def)
  case (Fun f ts)
  let ?I = "\<lambda> s. pi\<guillemotleft>\<alpha>,s\<guillemotright>"
  let ?E = "\<lambda> p. eval_lpoly \<alpha> p"
  show ?case 
  proof (cases "pi f (length ts)")
    case (Pair c cs)
    from wf_pi have "wf_elem (fst (pi f (length ts))) \<and> (\<forall> c \<in> set (snd (pi f (length ts))). wf_elem c \<and> geq c zero)" by auto
    with Pair have wf_c: "wf_elem c" and wf_cs: "\<forall> c \<in> set cs. wf_elem c" and pos_cs: "\<forall> c \<in> set cs. c \<ge> 0" by auto
    let ?plist = "map (\<lambda> at. mul_lpoly (fst at) (Pright (snd at))) (zip cs ts)"
    have wf_ps: "\<forall> p \<in> set ?plist. wf_lpoly p"
    proof
      fix p
      assume "p \<in> set ?plist"
      from this obtain a t where p: "p = mul_lpoly a (Pright t)" and at: "(a,t) \<in> set (zip cs ts)" by auto 
      from at have a: "a \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = a and y = t] set_zip_rightD[where x = a and y = t] by auto
      show "wf_lpoly p" by (simp only: p, rule wf_mul_lpoly, auto simp: wf_cs a t Fun)
    qed
    have wf_sum: "wf_lpoly (list_sum_lpoly ?plist)"
      by (rule wf_list_sum_lpoly, rule wf_ps)

    have pos_ps: "\<forall> p \<in> set ?plist. pos_coeffs p"
    proof
      fix p
      assume "p \<in> set ?plist"
      from this obtain a t where p: "p = mul_lpoly a (Pright t)" and at: "(a,t) \<in> set (zip cs ts)" by auto 
      from at have a: "a \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = a and y = t] set_zip_rightD[where x = a and y = t] by auto
      show "pos_coeffs p" by (simp only: p, rule pos_mul_lpoly, auto simp: a pos_cs wf_cs t Fun)
    qed
    have pos_sum: "pos_coeffs (list_sum_lpoly ?plist)"
      by (rule pos_list_sum_lpoly, rule wf_ps, rule pos_ps)

    have id1: "?E (list_sum_lpoly ?plist) = list_sum (map ?E ?plist)" by (rule list_sum_lpoly_sound, rule wf_ps, rule wf_ass)
    have "\<dots> \<ge> (list_sum (map (\<lambda> at. fst at * (?I (snd at))) (zip cs ts)))" (is "geq _ (list_sum ?vlist)")
    proof (simp only: o_def map_compose[symmetric], rule list_sum_mono, rule ballI)
      fix ct
      assume mem: "ct \<in> set (zip cs ts)"
      from this obtain c t where ct: "ct = (c,t)" by force
      from ct mem have c: "c \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = c and y = t] set_zip_rightD[where x = c and y = t] by auto
      have "(?E (mul_lpoly c (Pright t)) \<ge> (c * (?I t))) \<and> wf_elem (c * ?I t) \<and> wf_elem (?E (mul_lpoly c (Pright t)))" 
      proof (rule conjI)
	have id: "?E (mul_lpoly c (Pright t)) = c * (?E (Pright t))" by (rule mul_poly_sound, auto simp: c wf_cs Fun t wf_ass)
	show "?E (mul_lpoly c (Pright t)) \<ge> c * (?I t)" 
	  by (simp only: id, rule times_right_mono, auto simp: c wf_cs Fun  pos_cs t Fun wf_ass, rule wf_eval_lpoly, auto simp: wf_ass t Fun)
      next
	show "wf_elem (c * (?I t)) \<and> wf_elem (?E (mul_lpoly c (Pright t)))"
	by (rule conjI, rule wf_times, auto simp: c wf_cs wf_ass, rule wf_eval_lpoly, auto simp: wf_ass t Fun, rule wf_mul_lpoly, auto simp: c wf_cs t Fun)
      qed	
      with ct show " (eval_lpoly \<alpha> (mul_lpoly (fst ct) (Pright (snd ct))) \<ge> fst ct * (pi\<guillemotleft>\<alpha>,snd ct\<guillemotright>)) \<and> wf_elem (eval_lpoly \<alpha> (mul_lpoly (fst ct) (Pright (snd ct)))) \<and> wf_elem (fst ct * (pi\<guillemotleft>\<alpha>,snd ct\<guillemotright>))" by simp
    qed
    hence part1: "?E (list_sum_lpoly ?plist) \<ge> list_sum ?vlist" by (simp only: id1)  

    have wf_csump: "wf_lpoly (sum_lpoly (c_lpoly c) (list_sum_lpoly ?plist))" (is "wf_lpoly ?csump")
      by (rule wf_sum_lpoly, auto simp: wf_c wf_pvars_def, rule wf_sum)
    have pos_csump: "pos_coeffs ?csump"
      by (rule pos_sum_lpoly, auto simp: wf_sum wf_c pos_sum, auto simp: wf_c wf_pvars_def pos_coeffs_def pos_pvars_def)
    from Pair Fun have id: "Pright (Fun f ts) = (case ?csump of LPoly d nc \<Rightarrow> LPoly (max0 d) nc)" by auto
    
    from wf_csump have wf_final: "wf_lpoly (Pright (Fun f ts))" by (simp only: id, cases ?csump, auto simp: wf_pvars_def wf_max0)
    from pos_csump have pos_final: "pos_coeffs (Pright (Fun f ts))" unfolding pos_coeffs_def by (simp only: id, cases ?csump, auto)

    have "?E ?csump = ?E (c_lpoly c) + ?E (list_sum_lpoly ?plist)" by (rule sum_poly_sound, auto simp: wf_ass wf_c wf_pvars_def wf_sum)
    also have "\<dots> = c + ?E (list_sum_lpoly ?plist)" by (simp add: wf_c)
    finally have idr: "?E ?csump = c + ?E (list_sum_lpoly ?plist)" .
    
    have wf_sumr: "wf_elem (list_sum ?vlist)"
    proof (rule wf_list_sum, rule ballI)
      fix v
      assume "v \<in> set ?vlist"
      from this obtain c t where v: "v = c * (?I t)" and ct: "(c,t) \<in> set (zip cs ts)" by auto 
      from ct have c: "c \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = c and y = t] set_zip_rightD[where x = c and y = t] by auto
      show "wf_elem v" by (simp only: v, rule wf_times, auto simp: wf_cs c wf_ass)
    qed

    have ge_csum: "?E ?csump \<ge> c + list_sum ?vlist" 
      by (simp only: idr, rule plus_right_mono, rule part1, rule wf_c,
      rule wf_eval_lpoly, rule wf_ass, rule wf_sum, rule wf_sumr)

    have wf_csuml: "wf_elem (c + list_sum ?vlist)" using wf_sumr wf_c by auto

    from Pair have right: "?I (Fun f ts) = max0 ( c + list_sum ?vlist)" by auto 
    have "max0 ( ?E ?csump) \<ge> \<dots>"
      by (rule max0_mono, rule ge_csum, rule wf_eval_lpoly, rule wf_ass, rule wf_csump, rule wf_csuml)
    hence part2: "max0 ( ?E ?csump) \<ge> ?I (Fun f ts)" by (simp only: right)
    
    have part3: "?E (Pright (Fun f ts)) \<ge> max0 (?E ?csump)"
    proof (cases ?csump)
      case (LPoly d nc)
      have wf_nc: "wf_pvars nc" using wf_csump LPoly by (simp only: id, auto)
      have wf_enc: "wf_elem (eval_pvars \<alpha> nc)" using wf_csump LPoly by (simp only: id, auto simp: wf_ass)
      have wf_d: "wf_elem d" using wf_csump LPoly by (simp only: id, auto)
      have pos_nc: "pos_pvars nc" using pos_csump LPoly unfolding pos_coeffs_def by (simp only: id, auto)
      have nc_pos: "eval_pvars \<alpha> nc \<ge> 0" by (rule pos_eval_pvars, auto simp: wf_ass pos_ass pos_nc wf_nc)
      have md_pos: "max0 d \<ge> 0" by (rule max0_pos, rule wf_d)
      have "max0 d + eval_pvars \<alpha> nc \<ge> 0 + 0" (is "geq ?pr _") by (rule plus_left_right_mono, rule md_pos, rule nc_pos, auto simp: wf_enc wf_d wf_max0)
      hence "?pr \<ge> 0" by simp
      hence max_id: "max0 (?pr) = ?pr" by (rule max0_id_pos, auto simp: wf_enc wf_d wf_max0)
      show ?thesis by (simp only: id, simp only: LPoly, simp, rule geq_trans[where ?y = "max0 (?pr)"], simp only: max_id, rule geq_refl, auto simp: wf_enc wf_d wf_max0,
	rule max0_mono, rule plus_left_mono, rule max0_x, auto simp: wf_enc wf_d wf_max0)
    qed

    have ge_final: "?E (Pright (Fun f ts)) \<ge> ?I (Fun f ts)"
      by (rule geq_trans, rule part3, rule part2, rule wf_eval_lpoly, rule wf_ass, rule wf_final, rule wf_max0,
	rule wf_eval_lpoly, rule wf_ass, rule wf_csump, rule wf_terms, rule wf_ass)
    
    from wf_final ge_final pos_final show ?thesis by auto
  qed
qed


lemma Pright_sound :   assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "eval_lpoly \<alpha> (Pright t) \<ge> (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
  using Pright_both assms by auto

lemma wf_Pright[simp] : "wf_lpoly (Pright t)"
  using Pright_both[where \<alpha> = zero_ass] by auto


abbreviation check_polo_s :: "('f,'v :: sshow)rule \<Rightarrow> string list monad"
where "check_polo_s \<equiv> check_polo_sI plus_single_mono zero one plus times max0 geq gt pi" 

abbreviation check_polo_ns :: "('f,'v :: sshow)rule \<Rightarrow> string list monad"
where "check_polo_ns \<equiv> check_polo_nsI zero one plus times max0 geq pi"

lemma check_polo_s_sound: assumes check: "isOK(check_polo_s (s,t))"
  shows "(s,t) \<in> inter_s"
unfolding inter_s_def
proof (simp, rule allI)
  fix \<alpha>
  show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)" (is "?ass \<longrightarrow> ?goal")
  proof
    assume ?ass
    from check have check1: "isOK(check_lpoly_s one (Pleft s) (Pright t))" by (auto simp: Let_def)
    have strict: "(Pleft s, Pright t) \<in> poly_s" 
      by (rule check_lpoly_s_sound, rule check1, auto)
    have ge1: "pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> eval_lpoly \<alpha> (Pleft s)" by (rule Pleft_sound, auto simp: `?ass`)
    have gt: "\<dots> > eval_lpoly \<alpha> (Pright t)" using strict unfolding poly_s_def using `?ass` by auto
    have ge2: "\<dots> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>" by (rule Pright_sound, auto simp: `?ass`)
    show ?goal by (rule compat, rule ge1, rule compat2, rule gt, rule ge2, auto simp: `?ass` wf_eval_lpoly)
  qed
qed

lemma check_polo_ns_sound: assumes check: "isOK(check_polo_ns (s,t))"
  shows "(s,t) \<in> inter_ns"
unfolding inter_ns_def
proof (simp, rule allI)
  fix \<alpha>
  show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)" (is "?ass \<longrightarrow> ?goal")
  proof
    assume ?ass
    from check have check1: "isOK(check_lpoly_ns one (Pleft s) (Pright t))" by (auto simp: Let_def)
    have strict: "(Pleft s, Pright t) \<in> poly_ns" 
      by (rule check_lpoly_ns_sound, rule check1, auto)
    have ge1: "pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> eval_lpoly \<alpha> (Pleft s)" by (rule Pleft_sound, auto simp: `?ass`)
    have gt: "\<dots> \<ge> eval_lpoly \<alpha> (Pright t)" using strict unfolding poly_ns_def using `?ass` by auto
    have ge2: "\<dots> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>" by (rule Pright_sound, auto simp: `?ass`)
    show ?goal by (rule geq_trans, rule ge1, rule geq_trans, rule gt, rule ge2, auto simp: `?ass` wf_eval_lpoly)
  qed
qed

lemma inter_redpair_impl: "redpair_impl inter_s inter_ns check_polo_s check_polo_ns"
  by (rule redpair_impl.intro, rule inter_redpair, unfold redpair_impl_axioms_def, auto simp: check_polo_ns_sound check_polo_s_sound) 
end

types ('f,'a)lpoly_interL = "(('f \<times> nat) \<times> ('a \<times> 'a list)) list"

(* default interpretation f/n \<rightarrow> 1 + x_1 + ... + x_n *)
fun to_lpoly_inter :: "'a \<Rightarrow> ('f,'a)lpoly_interL \<Rightarrow> ('f,'a)lpoly_inter"
where "to_lpoly_inter on I f n = (case lookup (f,n) I of None \<Rightarrow> (on,replicate n on) | Some cas \<Rightarrow> cas)"


(* create an arity that is not used *)
fun create_arity :: "('f,'a)lpoly_interL \<Rightarrow> nat"
  where "create_arity fcns = Suc (
            foldr (\<lambda> (fn,ccs) m. max m (snd fn)) fcns 0
            )"

lemma create_arity_sound : assumes
  m: "m \<ge> create_arity I"
  shows "to_lpoly_inter on I f m = (on, replicate m on)" (is "?I f m = _")
proof (cases "lookup (f,m) I")
  case None thus ?thesis by auto
next
  case (Some val)
  hence one: "((f,m),val) \<in> set I" by (rule lookup_sound)
  hence "m \<le> foldr (\<lambda> (fn, ccs) n. max n (snd fn)) I 0" 
  proof (induct I, simp)
    case (Cons fnccs I)
    thus ?case by  (cases fnccs, cases "snd fnccs", auto)
  qed
  with m show ?thesis by auto
qed

declare create_arity.simps [simp del]

lemma replicate_prop: assumes "p x"
  shows "\<forall> y \<in> set (replicate n x). p y"
using assms by (induct n, auto)

context poly_order
begin

lemma inter_ns_ce_compat: assumes pi: "pi = to_lpoly_inter one I" (is "_ = ?pi")
  shows "ceTrs (c, create_arity I) \<subseteq> inter_ns"
unfolding inter_ns_def
proof (rule subsetI, simp add: ceTrs.simps)
  fix lr
  let ?n = "create_arity I"
  let ?m = "Suc (Suc ?n)"
  have "?m \<ge> ?n" by auto
  hence inter: "pi c ?m = (one, replicate ?m one)" by (simp only: pi, rule create_arity_sound)

  let ?sum = "list_sum (replicate ?n (1 * (\<lambda>(a, as). max0 (a + 0)) (pi c 0)))"
  obtain a as where pi: "pi c 0 = (a,as)" by force
  have "wf_elem (fst (pi c 0))" using wf_pi by auto
  with pi have wf_a: "wf_elem a" by auto
  obtain b where b: "b = max0 (a + 0)" by auto
  have idb: "1 * (\<lambda>(a, as). max0 (a + 0)) (pi c 0) = b"
    by (simp add: pi b wf_a wf_max0)
  have wf_pos_b: "wf_elem b \<and> geq b 0" by (simp add: b, auto simp: wf_a wf_max0 max0_pos)
  have wf_sum: "wf_elem ?sum"
    by (simp only: idb, rule wf_list_sum, rule replicate_prop, auto simp: wf_pos_b)
  have pos_sum: "?sum \<ge> 0"
    by (simp only: idb, rule pos_list_sum, rule replicate_prop, auto simp: wf_pos_b)
  from wf_sum pos_sum obtain d where d: "?sum = d" and wf_pos_d: "wf_elem d \<and> geq d 0" by force

  fix x
  assume " (\<exists>t s. x =
               (Fun c (t # s # replicate ?n (Fun c [])), t)) \<or>
        (\<exists>t s. x =
               (Fun c (t # s # replicate ?n (Fun c [])), s))" ( is "?tc \<or> ?sc")
  thus "(\<lambda> (s,t).  \<forall>\<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)) x"
  proof
    assume ?tc
    from this obtain t s where id: "x = (Fun c (t # s # replicate ?n (Fun c [])), t)" (is "x = (?l, _)") by auto    
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)" 
    proof (intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)" (is "?wf_pos \<longrightarrow> ?goal")
      proof
	assume wf_pos_ass: ?wf_pos
	thus ?goal using inter 
	proof (auto simp: d wf_pos_d wf_max0)
	  let ?term = "((1 + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + d"
	  have max: "max0 ?term = ?term" by (rule max0_id_pos, auto simp: wf_pos_ass wf_pos_d, (rule sum_pos)+, auto simp: wf_pos_ass wf_pos_d)
	  have "max0 ?term \<ge> ((0 + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + 0) + 0"
	    by (simp only: max, (rule plus_left_right_mono)+, auto simp: wf_pos_ass wf_pos_d)
	  thus "max0 ?term \<ge> (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" by (auto simp: wf_pos_ass)
	qed
      qed
    qed
    thus ?thesis using id by simp
  next
    assume ?sc
    from this obtain t s where id: "x = (Fun c (t # s # replicate ?n (Fun c [])), s)" (is "x = (?l, _)") by auto    
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,s\<guillemotright>)" 
    proof (intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,s\<guillemotright>)" (is "?wf_pos \<longrightarrow> ?goal")
      proof
	assume wf_pos_ass: ?wf_pos
	thus ?goal using inter 
	proof (auto simp: d wf_pos_d wf_max0)
	  let ?term = "((1 + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + d"
	  have max: "max0 ?term = ?term" by (rule max0_id_pos, auto simp: wf_pos_ass wf_pos_d, (rule sum_pos)+, auto simp: wf_pos_ass wf_pos_d)
	  have "max0 ?term \<ge> ((0 + 0) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + 0"
	    by (simp only: max, (rule plus_left_right_mono)+, auto simp: wf_pos_ass wf_pos_d)
	  thus "max0 ?term \<ge> (pi\<guillemotleft>\<alpha>,s\<guillemotright>)" by (auto simp: wf_pos_ass)
	qed
      qed
    qed
    thus ?thesis using id by simp
  qed
qed
end

fun create_afI :: "'a \<Rightarrow> ('f,'a)lpoly_interL \<Rightarrow> 'f af"
  where "create_afI ze [] f n i = True"
    |   "create_afI ze (((g,m),(c,coeffs)) # rest) f n i = (if f = g \<and> m = n then (i < length coeffs \<and> coeffs ! i \<noteq> ze) else create_afI ze rest f n i)"

lemma create_af_sound : assumes "\<not> create_afI ze I f n i"
  shows "length (snd (to_lpoly_inter on I f n)) \<le> i \<or> (snd (to_lpoly_inter on I f n)) ! i = ze"
using assms
proof (induct I, simp)
  case (Cons gmcas I)
  show ?case 
  proof (cases gmcas)
    case (fields gm c as)
    show ?thesis
    proof (cases gm)
      case (Pair g m)
      from prems show ?thesis by (cases "f = g \<and> m = n", auto)
    qed
  qed
qed
      
declare create_afI.simps[simp del]

context poly_order_carrier 
begin
abbreviation create_af :: "('f,'a)lpoly_interL \<Rightarrow> 'f af"
  where "create_af \<equiv> create_afI zero"
end

context poly_order
begin

lemma inter_ns_af_compat : assumes pi: "pi = to_lpoly_inter one I" (is "_ = ?pi")
  shows "af_compatible (create_af I) inter_ns"
proof (unfold af_compatible_def, (rule allI)+)
  fix f :: 'f and bef and s t :: "('f,'v)term" and aft
  let ?st = "Fun f (bef @ s # aft)"
  let ?tt = "Fun f (bef @ t # aft)"
  let ?pair = "(?st, ?tt)"
  let ?n = "Suc (length bef + length aft)"
  let ?i = "length bef"
  show "create_af I f ?n ?i \<or> ?pair \<in> inter_ns"
  proof (cases "create_af I f ?n ?i")
    case True thus ?thesis ..
  next
    case False
    hence "length (snd (?pi f ?n)) \<le> ?i \<or> snd (?pi f ?n) ! ?i = 0" by (rule create_af_sound)
    with pi have pi3: "length (snd (pi f ?n)) \<le> ?i \<or> snd (pi f ?n) ! ?i = 0" by simp
    obtain c cs where pi2: "pi f ?n = (c,cs)" by force
    with pi3 have short_zero: "length cs \<le> ?i \<or> cs ! ?i = 0" by auto
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> ((pi\<guillemotleft>\<alpha>,?st\<guillemotright>) \<ge> (pi\<guillemotleft>\<alpha>,?tt\<guillemotright>))" (is "\<forall> \<alpha>. ?wfpos \<alpha> \<longrightarrow> ?rel \<alpha>")
    proof
      fix \<alpha>
      show "?wfpos \<alpha> \<longrightarrow> ?rel \<alpha>"
      proof
	assume wfpos: "?wfpos \<alpha>"
	hence wf_ass: "wf_ass \<alpha>" ..
	let ?in = "\<lambda> t :: ('f,'v) term. (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
	let ?inter = "\<lambda> s bef cs. list_sum (map (\<lambda> at. fst at * (?in (snd at))) (zip cs (bef @ s # aft)))"
	let ?intre = "\<lambda> aft cs. list_sum (map (\<lambda> at. fst at * (?in (snd at))) (zip cs aft))"
	have int_s: "?in ?st = max0 (c + ?inter s bef cs)" by (auto simp: pi2)
	have int_t: "?in ?tt = max0 (c + ?inter t bef cs)" by (auto simp: pi2)	
	from short_zero have "?inter s bef cs = ?inter t bef cs" 
	proof (induct cs arbitrary: bef)
	  case Nil thus ?case by auto
	next
	  case (Cons c cs)
	  show ?case 
	  proof (cases bef)
	    case Nil
	    with Cons show ?thesis by (simp add: wf_ass)
	  next
	    case (Cons b befs)
	    from prems have "length cs \<le> length befs \<or> cs ! length befs = 0" by auto
	    with prems show ?thesis by auto
	  qed
	qed
	with int_s int_t have eq: "?in ?st = ?in ?tt" by auto
	show "?rel \<alpha>" by (simp only: eq, rule geq_refl, rule wf_terms, auto simp: wf_ass)
      qed
    qed
    hence "?pair \<in> inter_ns" unfolding inter_ns_def by auto
    thus ?thesis .. 
  qed
qed
end

fun check_poly_monoI :: "'a :: sshow => ('a => 'a => bool) => ('a \<Rightarrow> bool) \<Rightarrow> ('f :: sshow,'a)lpoly_interL \<Rightarrow> string list monad"
where "check_poly_monoI ze ge cm = checkAllM (\<lambda> ((f,n),(c,cs)). 
 (   checkB (ge c ze) [''constant part '' +++ sshow c +++ '' must be at least '' +++ sshow ze] 
  \<guillemotright> checkB (n \<le> length cs) [''the last argument is ignored'']
  \<guillemotright> checkAllM (\<lambda> d. checkB (cm d) [''coefficient '' +++ sshow d +++ '' is not allowed'' ]) cs )
 \<leftarrow> (\<lambda> s. (''problem with monotonicity due to interpretation of '' +++ sshow f +++ ''/''  +++ sshow n) # s)
)"


locale mono_poly_order_carrier = poly_order_carrier +
  fixes check_mono :: "'a \<Rightarrow> bool"
  assumes plus_single_mono: "plus_single_mono"
  and check_mono: "\<lbrakk>check_mono x; gt y z; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> gt (times x y) (times x z)" 
  and one_mono[simp]: "check_mono 1"
begin
abbreviation check_poly_mono :: "('f :: sshow,'a :: sshow)lpoly_interL \<Rightarrow> string list monad"
where "check_poly_mono \<equiv> check_poly_monoI 0 geq check_mono"

lemma plus_gt_right_mono: "\<lbrakk>y > z; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x + y > x + z"
  by (simp add: plus_comm[where x = x], rule plus_gt_left_mono, auto simp: plus_single_mono)

abbreviation pi_of where "pi_of \<equiv> to_lpoly_inter one"

lemma check_poly_mono_sound : assumes "isOK(check_poly_mono I)"
  shows "\<forall> f n. (fst (pi_of I f n) \<ge> 0) \<and> (length (snd (pi_of I f n)) \<ge> n \<and> (\<forall> c \<in> set (snd (pi_of I f n)). check_mono c))" (is "\<forall> f n. ?prop f n")
proof (intro allI)
  fix f n
  show "?prop f n" 
  proof (cases "lookup (f,n) I")
    case None
    hence id: "pi_of I f n = (one, replicate n 1)" by auto
    show ?thesis
      by (simp only: id, rule conjI, simp, rule conjI, simp, simp,rule replicate_prop, rule one_mono)
  next
    case (Some ccs)
    from this obtain c cs where lookup: "lookup (f,n) I = Some (c,cs)" by force
    hence "((f,n),(c,cs)) \<in> set I" by (rule lookup_sound)
    with assms lookup show "?prop f n" by auto
  qed
qed

declare check_poly_monoI.simps[simp del]
end

locale mono_poly_order = poly_order + mono_poly_order_carrier + 
  assumes pi_mono: "\<forall> f n. (fst (pi f n) \<ge> 0) \<and> (length (snd (pi f n)) \<ge> n \<and> (\<forall> c \<in> set (snd (pi f n)). check_mono c))"
begin
lemma inter_s_ce_compat: assumes pi: "pi = to_lpoly_inter one I" (is "_ = ?pi")
  shows "ceTrs (c, create_arity I) \<subseteq> inter_s"
unfolding inter_s_def
proof (rule subsetI, simp add: ceTrs.simps)
  fix lr
  let ?n = "create_arity I"
  let ?m = "Suc (Suc ?n)"
  have "?m \<ge> ?n" by auto
  hence inter: "pi c ?m = (one, replicate ?m one)" by (simp only: pi, rule create_arity_sound)

  let ?sum = "list_sum (replicate ?n (1 * (\<lambda>(a, as). max0 (a + 0)) (pi c 0)))"
  obtain a as where pi: "pi c 0 = (a,as)" by force
  have "wf_elem (fst (pi c 0))" using wf_pi by auto
  with pi have wf_a: "wf_elem a" by auto
  obtain b where b: "b = max0 (a + 0)" by auto
  have idb: "1 * (\<lambda>(a, as). max0 (a + 0)) (pi c 0) = b"
    by (simp add: pi b wf_a wf_max0)
  have wf_pos_b: "wf_elem b \<and> geq b 0" by (simp add: b, auto simp: wf_a wf_max0 max0_pos)
  have wf_sum: "wf_elem ?sum"
    by (simp only: idb, rule wf_list_sum, rule replicate_prop, auto simp: wf_pos_b)
  have pos_sum: "?sum \<ge> 0"
    by (simp only: idb, rule pos_list_sum, rule replicate_prop, auto simp: wf_pos_b)
  from wf_sum pos_sum obtain d where d: "?sum = d" and wf_pos_d: "wf_elem d \<and> geq d 0" by force

  fix x
  assume " (\<exists>t s. x =
               (Fun c (t # s # replicate ?n (Fun c [])), t)) \<or>
        (\<exists>t s. x =
               (Fun c (t # s # replicate ?n (Fun c [])), s))" ( is "?tc \<or> ?sc")
  thus "(\<lambda> (s,t).  \<forall>\<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)) x"
  proof
    assume ?tc
    from this obtain t s where id: "x = (Fun c (t # s # replicate ?n (Fun c [])), t)" (is "x = (?l, _)") by auto    
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)" 
    proof (intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)" (is "?wf_pos \<longrightarrow> ?goal")
      proof
	assume wf_pos_ass: ?wf_pos
	thus ?goal using inter 
	proof (auto simp: d wf_pos_d wf_max0)
	  let ?term = "((1 + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + d"
	  have max: "max0 ?term = ?term" by (rule max0_id_pos, auto simp: wf_pos_ass wf_pos_d, (rule sum_pos)+, auto simp: wf_pos_ass wf_pos_d)
	  have "max0 ?term > ((0 + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + 0) + 0"
	    by (simp only: max, (rule plus_gt_left_mono)+, auto simp: wf_pos_ass wf_pos_d one_gt_zero, (rule plus_single_mono)+)
	  thus "max0 ?term > (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" by (auto simp: wf_pos_ass)
	qed
      qed
    qed
    thus ?thesis using id by simp
  next
    assume ?sc
    from this obtain t s where id: "x = (Fun c (t # s # replicate ?n (Fun c [])), s)" (is "x = (?l, _)") by auto    
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> > pi\<guillemotleft>\<alpha>,s\<guillemotright>)" 
    proof (intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> > pi\<guillemotleft>\<alpha>,s\<guillemotright>)" (is "?wf_pos \<longrightarrow> ?goal")
      proof
	assume wf_pos_ass: ?wf_pos
	thus ?goal using inter 
	proof (auto simp: d wf_pos_d wf_max0)
	  let ?term = "((1 + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + d"
	  have max: "max0 ?term = ?term" by (rule max0_id_pos, auto simp: wf_pos_ass wf_pos_d, (rule sum_pos)+, auto simp: wf_pos_ass wf_pos_d)
	  have "max0 ?term > ((0 + 0) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + 0"
	    by (simp only: max, (rule plus_gt_left_mono)+, auto simp: wf_pos_ass wf_pos_d one_gt_zero, (rule plus_single_mono)+)
	  thus "max0 ?term > (pi\<guillemotleft>\<alpha>,s\<guillemotright>)" by (auto simp: wf_pos_ass)
	qed
      qed
    qed
    thus ?thesis using id by simp
  qed
qed

lemma inter_s_mono: "one_step_ctxt_closed inter_s" 
unfolding one_step_ctxt_closed_def
proof ((rule allI)+)
  fix f :: 'b and bef and s t :: "('b,'v)term" and aft
  let ?st = "Fun f (bef @ s # aft)"
  let ?tt = "Fun f (bef @ t # aft)"
  let ?pair = "(?st, ?tt)"
  let ?n = "Suc (length bef + length aft)"
  let ?i = "length bef"
  obtain c cs where ccs: "pi f ?n = (c,cs)" by force
  from pi_mono have "((fst (pi f ?n)) \<ge> 0) \<and> length (snd (pi f ?n)) \<ge> ?n \<and> (\<forall> c \<in> set (snd (pi f ?n)). check_mono c)" by auto
  with ccs have mono: "(c \<ge> 0) \<and> length cs \<ge> ?n \<and> (\<forall> d \<in> set cs. check_mono d)" by auto
  from wf_pi have "wf_elem (fst (pi f ?n))" and "\<forall> d \<in> set (snd (pi f ?n)). wf_elem d \<and> geq d 0" by auto
  with ccs have wf_c: "wf_elem c" and wf_cs: "\<forall> d \<in> set cs. wf_elem d \<and> geq d 0" by auto
  show "(s,t) \<in> inter_s \<longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> inter_s"
  proof
    assume st: "(s,t) \<in> inter_s"
    show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> inter_s"
      unfolding inter_s_def
    proof (clarify)
      fix \<alpha> :: "('v,'a)p_ass"
      assume wf_ass: "wf_ass \<alpha>" and pos_ass: "pos_ass \<alpha>"
      let ?in = "\<lambda> t :: ('b,'v) term. (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
      let ?inter = "\<lambda> s bef cs. list_sum (map (\<lambda> at. fst at * (?in (snd at))) (zip cs (bef @ s # aft)))"
      have int_s: "?in ?st = max0 (c + ?inter s bef cs)" by (auto simp: ccs)
      have int_t: "?in ?tt = max0 (c + ?inter t bef cs)" by (auto simp: ccs)	
      from mono wf_cs have main: "?inter s bef cs > ?inter t bef cs" 
      proof (induct cs arbitrary: bef)
	case Nil thus ?case by auto
      next
	case (Cons c cs)
	hence wf_c: "wf_elem c" by auto
	have wf_rec: "\<And> aft. wf_elem (list_sum (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip cs aft)))"
	proof (rule wf_list_sum, auto)
	  fix aft and a :: 'a and b :: "('b,'v)term"
	  assume "(a,b) \<in> set (zip cs aft)"
	  hence "a \<in> set cs" by (rule set_zip_leftD)
	  with Cons have "wf_elem a" by auto
	  thus "wf_elem (a * ?in b)" using wf_ass by auto
	qed	
	show ?case
	proof (cases bef)
	  case Nil
	  from st have "?in s > ?in t" unfolding inter_s_def using wf_ass pos_ass by auto
	  show ?thesis by (simp add: Nil Cons, rule plus_gt_left_mono, rule check_mono, auto simp: `?in s > ?in t` Cons wf_ass plus_single_mono, rule geq_refl,
	    (rule wf_rec)+)
	next
	  case (Cons b befs)
	  from prems have "Suc (length befs + length aft) \<le> length cs" by auto
	  with prems have rec: "?inter s befs cs > ?inter t befs cs" by auto
	  show ?thesis  by (simp add: prems, rule plus_gt_right_mono, rule rec, auto simp: wf_c wf_ass, (rule wf_rec)+)
	qed
      qed
      have wf_inter: "\<And> t. wf_elem (?inter t bef cs)"
      proof (rule wf_list_sum, auto)
	fix t a b
	assume "(a,b) \<in> set (zip cs (bef @ t # aft))"
	hence "a \<in> set cs" by (rule set_zip_leftD)
	with wf_cs wf_ass show "wf_elem (a * (?in b))" by auto
      qed	
      have wf_cint: "\<And> t. wf_elem (c + ?inter t bef cs)"
	by (rule wf_plus, rule wf_c, rule wf_inter)
      have pos_inter: "\<And> t. (?inter t bef cs) \<ge> 0"
      proof (rule pos_list_sum, (clarify, simp)+)
	fix t a b 
	assume "(a,b) \<in> set (zip cs (bef @ t # aft))"
	hence "a \<in> set cs" by (rule set_zip_leftD)
	hence pos_a: "a \<ge> 0" and wf_a: "wf_elem a" using wf_cs  by auto
	show "wf_elem (a * (?in b)) \<and> (a * (?in b) \<ge> 0)"
	proof (rule conjI, auto simp: wf_ass wf_a)
	  have int: "(a * (?in b) \<ge> a * 0)"
	    by (rule times_right_mono, rule pos_a, rule pos_term, auto simp: wf_a wf_ass pos_ass)
	  show "(a * (?in b)) \<ge> 0" by (rule geq_trans[where y = "a * 0"], rule int, auto simp: wf_a wf_ass)
	qed
      qed
      have "\<And> t. c + (?inter t bef cs) \<ge> 0 + 0"
	by (rule plus_left_right_mono, auto simp: mono, rule pos_inter, rule wf_c, rule wf_inter)
      hence pos_cint: "\<And> t. c + (?inter t bef cs) \<ge> 0" by auto
      have id_cint: "\<And> t. max0(c + (?inter t bef cs)) = c + (?inter t bef cs)"
	by (rule max0_id_pos, rule pos_cint, rule wf_cint)
      have wf_rec: "\<And> t aft. wf_elem (list_sum (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip cs (bef @ t # aft))))"
      proof (rule wf_list_sum, auto)
	fix t aft and a :: 'a and b :: "('b,'v)term"
	assume "(a,b) \<in> set (zip cs (bef @ t # aft))"
	hence "a \<in> set cs" by (rule set_zip_leftD)
	with wf_cs have "wf_elem a" by auto
	thus "wf_elem (a * ?in b)" using wf_ass by auto
      qed	
      show "?in ?st > ?in ?tt" by (simp only: int_s int_t id_cint, rule plus_gt_right_mono, rule main, rule wf_c, (rule wf_rec)+)
    qed
  qed
qed
end


fun check_lpoly_coeffs :: "('a :: sshow \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('f :: sshow,'a)lpoly_interL \<Rightarrow> string list monad"
where "check_lpoly_coeffs wfe ch I = (
  (checkAllM (\<lambda> ((f,n),(c,cs)). 
    (checkB (wfe c) [''constant part '' +++ sshow c +++ '' is not well-formed'']
  \<guillemotright> checkAllM (\<lambda> a. checkB (ch a \<and> wfe a) [''coefficient '' +++ sshow a +++ '' is not allowed'']) cs)
  \<leftarrow> (\<lambda> s. (''problem with interpretation of '' +++ sshow f +++ ''/'' +++ sshow n) # s)) I))"

context poly_order_carrier
begin
lemma check_lpoly_coeffs_sound: assumes check: "isOK(check_lpoly_coeffs wf_elem (\<lambda> x. geq x 0) I)"
  shows "poly_order wf_elem zero plus times geq gt plus_single_mono one max0 (to_lpoly_inter one I)"  
proof
  show "wf_lpoly_inter (to_lpoly_inter one I)" (is "wf_lpoly_inter ?I")
  proof (rule allI, rule allI)
    fix f n 
    let ?goal = "\<lambda> f n. wf_elem (fst (?I f n)) \<and> (\<forall>a. a \<in> set (snd (?I f n)) \<longrightarrow> wf_elem a \<and> (a \<ge> 0))"
    show "?goal f n"
    proof (cases "lookup (f,n) I")
      case None
      hence repl: "?I f n = (1, replicate n 1)" by auto
      have "\<forall> x \<in> set (replicate n 1). x = one" by (induct n, auto)
      with repl show "?goal f n" by auto
    next
      case (Some aas)
      hence aas: "?I f n = aas" by auto
      from Some have "((f,n),aas) \<in> set I" by (rule lookup_sound)
      with check have "wf_elem (fst aas) \<and> (\<forall> a \<in> set (snd aas). wf_elem a \<and> geq a 0)" by (cases aas, auto)
      with aas show "?goal f n" by auto
    qed
  qed
qed
end

types 'a poly_components = "(('a \<Rightarrow> bool) \<times> 'a \<times> 'a \<times> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<times> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<times> ('a \<Rightarrow> 'a) \<times> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<times> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<times> bool                \<times> string)" 
(*                           wf_elem         0   1     plus               times             max0         \<ge>                    >                plus_single_mono      description*)

fun get_comp_wf_elem :: "'a poly_components \<Rightarrow> ('a \<Rightarrow> bool)"
where "get_comp_wf_elem (wfe,ze,on,pl,ti,mz,ge,gt,mo,de) = wfe"

fun get_comp_0 :: "'a poly_components \<Rightarrow> 'a"
where "get_comp_0 (wfe,ze,on,pl,ti,mz,ge,gt,mo,de) = ze"

fun get_comp_1 :: "'a poly_components \<Rightarrow> 'a"
where "get_comp_1 (wfe,ze,on,pl,ti,mz,ge,gt,mo,de) = on"

fun get_comp_plus :: "'a poly_components \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a)"
where "get_comp_plus (wfe,ze,on,pl,ti,mz,ge,gt,mo,de) = pl"

fun get_comp_times :: "'a poly_components \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a)"
where "get_comp_times (wfe,ze,on,pl,ti,mz,ge,gt,mo,de) = ti"

fun get_comp_max0 :: "'a poly_components \<Rightarrow> ('a \<Rightarrow> 'a)"
where "get_comp_max0 (wfe,ze,on,pl,ti,mz,ge,gt,mo,de) = mz"

fun get_comp_ge :: "'a poly_components \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool)"
where "get_comp_ge (wfe,ze,on,pl,ti,mz,ge,gt,mo,de) = ge"

fun get_comp_gt :: "'a poly_components \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool)"
where "get_comp_gt (wfe,ze,on,pl,ti,mz,ge,gt,mo,de) = gt"

fun get_comp_mono :: "'a poly_components \<Rightarrow> bool"
where "get_comp_mono (wfe,ze,on,pl,ti,mz,ge,gt,mo,de) = mo"

fun get_comp_description :: "'a poly_components \<Rightarrow> string"
where "get_comp_description (wfe,ze,on,pl,ti,mz,ge,gt,mo,de) = de"

abbreviation check_polo_s_comp :: "('a :: sshow) poly_components \<Rightarrow> ('f :: sshow,'a)lpoly_inter \<Rightarrow> ('f,'v :: sshow)rule \<Rightarrow> string list monad"
where "check_polo_s_comp c \<equiv> check_polo_sI (get_comp_mono c) (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c) (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c)"

abbreviation check_polo_ns_comp :: "('a :: sshow) poly_components \<Rightarrow> ('f :: sshow,'a)lpoly_inter \<Rightarrow> ('f,'v :: sshow)rule \<Rightarrow> string list monad"
where "check_polo_ns_comp c \<equiv> check_polo_nsI (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c) (get_comp_max0 c) (get_comp_ge c)"

abbreviation check_polo_coeffs_comp :: "('a :: sshow) poly_components \<Rightarrow> ('f :: sshow,'a)lpoly_interL \<Rightarrow> string list monad"
where "check_polo_coeffs_comp c \<equiv> check_lpoly_coeffs (get_comp_wf_elem c) (\<lambda> x. (get_comp_ge c) x (get_comp_0 c))"


context poly_order_carrier
begin
abbreviation carrier_components :: "'a poly_components"
where "carrier_components \<equiv> (wf_elem, zero, one, plus, times, max0, geq, gt, plus_single_mono, description)"

lemma conditional_poly_order: assumes condition: "isOK(check_polo_coeffs_comp carrier_components I)"
  shows "poly_order wf_elem zero plus times geq gt plus_single_mono one max0 (to_lpoly_inter one I)"
proof (rule check_lpoly_coeffs_sound)
  show "isOK (check_lpoly_coeffs wf_elem (\<lambda> x. x \<ge> 0) I)"
    using condition by auto
qed

lemma conditional_inter_redpair: assumes condition: "isOK(check_polo_coeffs_comp carrier_components I)"
  shows "redpair
  (poly_order.inter_s wf_elem zero plus times geq gt max0 (to_lpoly_inter one I)) 
  (poly_order.inter_ns wf_elem zero plus times geq max0 (to_lpoly_inter one I))"
  by (rule poly_order.inter_redpair, rule conditional_poly_order, rule condition) 

lemma conditional_inter_redpair_impl: assumes condition: "isOK(check_polo_coeffs_comp carrier_components I)"
  shows "redpair_impl 
  (poly_order.inter_s wf_elem zero plus times geq gt max0 (to_lpoly_inter one I)) 
  (poly_order.inter_ns wf_elem zero plus times geq max0 (to_lpoly_inter one I))
  (check_polo_s_comp carrier_components (to_lpoly_inter one I))
  (check_polo_ns_comp carrier_components (to_lpoly_inter one I))"
proof -
  have "redpair_impl 
    (poly_order.inter_s wf_elem zero plus times geq gt max0 (to_lpoly_inter one I))
    (poly_order.inter_ns wf_elem zero plus times geq max0 (to_lpoly_inter one I))
    (check_polo_sI plus_single_mono zero one plus times max0 geq gt (to_lpoly_inter one I))
    (check_polo_nsI zero one plus times max0 geq (to_lpoly_inter one I))"    
  proof (rule poly_order.inter_redpair_impl)
    show "poly_order wf_elem 0 plus times geq gt plus_single_mono 1 max0 (to_lpoly_inter 1 I)"
    proof (rule check_lpoly_coeffs_sound)
      have "(check_lpoly_coeffs wf_elem (\<lambda> x. x \<ge> 0) I) = (check_polo_coeffs_comp carrier_components I)" by auto
      with condition show "isOK(check_lpoly_coeffs wf_elem (\<lambda> x. x \<ge> 0) I)" by auto
    qed
  qed
  thus ?thesis by auto
qed

lemma conditional_inter_af_redpair: assumes condition: "isOK(check_polo_coeffs_comp carrier_components I)"
  shows "af_redpair
  (poly_order.inter_s wf_elem zero plus times geq gt max0 (to_lpoly_inter one I)) 
  (poly_order.inter_ns wf_elem zero plus times geq max0 (to_lpoly_inter one I))
  (create_af I)"
  by (rule af_redpair.intro, rule conditional_inter_redpair, rule condition, unfold af_redpair_axioms_def,
  rule poly_order.inter_ns_af_compat, rule conditional_poly_order, rule condition, simp)

lemma conditional_inter_ce_redpair: assumes condition: "isOK(check_polo_coeffs_comp carrier_components I)"
  shows "ce_redpair
  (poly_order.inter_s wf_elem zero plus times geq gt max0 (to_lpoly_inter one I)) 
  (poly_order.inter_ns wf_elem zero plus times geq max0 (to_lpoly_inter one I))
  (create_arity I)
  c
  "
  by (rule ce_redpair.intro, rule conditional_inter_redpair, rule condition, unfold ce_redpair_axioms_def, 
      rule poly_order.inter_ns_ce_compat, rule conditional_poly_order, rule condition, auto)

lemma conditional_inter_ce_redpair_impl: assumes condition: "isOK(check_polo_coeffs_comp carrier_components I)"
  shows "ce_redpair_impl 
  (poly_order.inter_s wf_elem zero plus times geq gt max0 (to_lpoly_inter one I)) 
  (poly_order.inter_ns wf_elem zero plus times geq max0 (to_lpoly_inter one I))
  (check_polo_s_comp carrier_components (to_lpoly_inter one I))
  (check_polo_ns_comp carrier_components (to_lpoly_inter one I))
  (create_arity I)
  c
  "
  by (rule ce_redpair_impl.intro, rule conditional_inter_redpair_impl, rule condition, 
    rule conditional_inter_ce_redpair, rule condition)

lemma conditional_inter_ce_af_redpair_impl: assumes condition: "isOK(check_polo_coeffs_comp carrier_components I)"
  shows "ce_af_redpair_impl 
  (poly_order.inter_s wf_elem zero plus times geq gt max0 (to_lpoly_inter one I)) 
  (poly_order.inter_ns wf_elem zero plus times geq max0 (to_lpoly_inter one I))
  (check_polo_s_comp carrier_components (to_lpoly_inter one I))
  (check_polo_ns_comp carrier_components (to_lpoly_inter one I))
  (create_arity I)
  c
  (create_af I)
  "
  by (rule ce_af_redpair_impl.intro, rule conditional_inter_redpair_impl, rule condition, rule ce_af_redpair.intro,
    rule conditional_inter_ce_redpair, rule condition, rule conditional_inter_af_redpair, rule condition)
end

context mono_poly_order_carrier
begin

lemma conditional_inter_mono_poly_order: assumes condition: "isOK(check_polo_coeffs_comp carrier_components I)"
  and condition2: "isOK(check_poly_mono I)"
  shows "mono_poly_order wf_elem zero plus times geq gt plus_single_mono one max0 (to_lpoly_inter one I) check_mono"
by (rule mono_poly_order.intro,  rule conditional_poly_order, rule condition, rule mono_poly_order_carrier_axioms, 
    unfold mono_poly_order_axioms_def, rule check_poly_mono_sound, rule condition2)


lemma conditional_inter_mono_redpair_impl: assumes condition: "isOK(check_polo_coeffs_comp carrier_components I)"
  shows "mono_redpair_impl 
  (poly_order.inter_s wf_elem zero plus times geq gt max0 (to_lpoly_inter one I)) 
  (poly_order.inter_ns wf_elem zero plus times geq max0 (to_lpoly_inter one I))
  (check_polo_s_comp carrier_components (to_lpoly_inter one I))
  (check_polo_ns_comp carrier_components (to_lpoly_inter one I))
  (check_poly_mono I)"
by (rule mono_redpair_impl.intro, rule conditional_inter_redpair_impl, rule condition, unfold mono_redpair_impl_axioms_def, rule impI,
    rule mono_poly_order.inter_s_mono, rule conditional_inter_mono_poly_order, rule condition, blast)


lemma conditional_inter_mono_ce_redpair: assumes condition: "isOK(check_polo_coeffs_comp carrier_components I)"
  and condition2: "isOK(check_poly_mono I)"
  shows "mono_ce_redpair
  (poly_order.inter_s wf_elem zero plus times geq gt max0 (to_lpoly_inter one I)) 
  (poly_order.inter_ns wf_elem zero plus times geq max0 (to_lpoly_inter one I))
  (create_arity I) c"
by (rule mono_ce_redpair.intro, rule conditional_inter_ce_redpair, rule condition, unfold mono_ce_redpair_axioms_def, rule mono_poly_order.inter_s_ce_compat,
  rule conditional_inter_mono_poly_order, rule condition, rule condition2, simp)

lemma conditional_inter_ce_mono_redpair_impl: assumes condition: "isOK(check_polo_coeffs_comp carrier_components I)"
  and condition2: "isOK(check_poly_mono I)"
  shows "ce_mono_redpair_impl
  (poly_order.inter_s wf_elem zero plus times geq gt max0 (to_lpoly_inter one I)) 
  (poly_order.inter_ns wf_elem zero plus times geq max0 (to_lpoly_inter one I))
  (create_arity I) c
  (check_polo_s_comp carrier_components (to_lpoly_inter one I))
  (check_polo_ns_comp carrier_components (to_lpoly_inter one I))
  (check_poly_mono I)"
by (rule ce_mono_redpair_impl.intro, rule conditional_inter_mono_ce_redpair, rule condition, rule condition2, 
    rule conditional_inter_mono_redpair_impl, rule condition)
end

fun create_lpoly_repr :: "'a poly_components \<Rightarrow> ('f :: sshow, 'a :: sshow)lpoly_interL \<Rightarrow> string list"
where "create_lpoly_repr c I = (let pi = to_lpoly_inter (get_comp_1 c) I in  (
  [''polynomial interpretration over '' +++ get_comp_description c] @ (map 
  (\<lambda> (f,n). let t = Fun f (map Var (fresh_strings ''x_'' 1 n [])) in (''Pol('' +++ sshow t +++ '') = '' +++
  show_lpoly (get_comp_0 c) (get_comp_1 c) (PleftI (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c) (get_comp_max0 c) pi t)
)) (remdups (map fst I)))))"  
  

fun create_rule_removal_proc_comp :: "('a :: sshow)poly_components \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('f :: sshow,'a)lpoly_interL \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_rule_removal_proc_comp c cm I R rR  = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  checkRuleRemovalProc (check_poly_monoI (get_comp_0 c) (get_comp_ge c) cm I)
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +++ get_comp_description c +++ '')'') 
           (create_lpoly_repr c I)
           R rR))"

fun create_redpair_proc_comp :: "('a :: sshow)poly_components \<Rightarrow> ('f :: sshow,'a)lpoly_interL \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_redpair_proc_comp c I P rP R = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  checkRedPairProc 
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +++ get_comp_description c +++ '')'') 
           (create_lpoly_repr c I)
           P rP R))"


fun create_mono_redpair_proc_comp :: "('a :: sshow)poly_components \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('f :: sshow,'a)lpoly_interL \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
  where "create_mono_redpair_proc_comp c cm I P rP R rR = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  checkMonoRedPairProc (check_poly_monoI (get_comp_0 c) (get_comp_ge c) cm I)
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +++ get_comp_description c +++ '')'') 
           (create_lpoly_repr c I)
           P rP R rR))"

fun create_redpair_ur_proc_comp :: "('a :: sshow)poly_components \<Rightarrow> ('f :: sshow,'a)lpoly_interL \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_redpair_ur_proc_comp c I P rP R rm ur = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  checkRedPairUrProc 
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +++ get_comp_description c +++ '')'') 
           (create_lpoly_repr c I)
           P rP R rm ur))"

fun create_mono_redpair_ur_proc_comp :: "('a :: sshow)poly_components \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('f :: sshow,'a)lpoly_interL \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_mono_redpair_ur_proc_comp c cm I P rP R rm rR ur = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  checkMonoRedPairUrProc (check_poly_monoI (get_comp_0 c) (get_comp_ge c) cm I)
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +++ get_comp_description c +++ '')'') 
           (create_lpoly_repr c I)
           P rP R rm rR ur))"

fun create_redpair_ur_af_proc_comp :: "('a :: sshow)poly_components \<Rightarrow> ('f :: sshow,'a)lpoly_interL \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_redpair_ur_af_proc_comp c I P rP R rm ur = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  checkRedPairUrAfProc 
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +++ get_comp_description c +++ '')'') 
           (create_lpoly_repr c I)
           P rP R rm ur
           (create_afI (get_comp_0 c) I)))"



context poly_order_carrier
begin
lemma redpair_proc_comp_ci : 
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp carrier_components I P rP R))"
unfolding "chain_identifying_onlyDP_proc_impl_def"
proof (intro allI, intro impI)
  fix P R rP
  let ?c = carrier_components
  let ?pi = "to_lpoly_inter one I"
  assume check: "isOK(create_redpair_proc_comp ?c I P rP R)"
  hence check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def)
  from check have accept: "isOK(checkRedPairProc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) [] [] P rP R)" (is "isOK(?proc)") 
    by (auto simp: Let_def) 
  have "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(checkRedPairProc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) [] [] P rP R))"
    by (rule redpair_impl.check_redpair_proc_sound, rule conditional_inter_redpair_impl, rule check1)
  thus "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
    unfolding chain_identifying_onlyDP_proc_impl_def
    using accept by best
qed

lemma redpair_ur_proc_comp_ci : 
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp carrier_components I P rP R rm ur))"
unfolding "chain_identifying_onlyDP_proc_impl_def"
proof (intro allI, intro impI)
  fix P R rP
  let ?c = carrier_components
  let ?pi = "to_lpoly_inter one I"
  assume check: "isOK(create_redpair_ur_proc_comp ?c I P rP R (computeRuleMap R) ur)"
  hence check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def)
  from check have accept: "isOK(checkRedPairUrProc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) [] [] P rP R (computeRuleMap R) ur)" (is "isOK(?proc)") 
    by (auto simp: Let_def) 
  have "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(checkRedPairUrProc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) [] [] P rP R rm ur))"
    by (rule ce_redpair_impl.check_redpair_ur_proc_sound, rule conditional_inter_ce_redpair_impl, rule check1)
  thus "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
    unfolding chain_identifying_onlyDP_proc_impl_def
    using accept by best
qed

lemma redpair_ur_af_proc_comp_ci : 
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp carrier_components I P rP R rm ur))"
unfolding "chain_identifying_onlyDP_proc_impl_def"
proof (intro allI, intro impI)
  fix P R rP
  let ?c = carrier_components
  let ?pi = "to_lpoly_inter one I"
  assume check: "isOK(create_redpair_ur_af_proc_comp ?c I P rP R (computeRuleMap R) ur)"
  hence check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def)
  let ?af = "create_af I"
  from check have accept: "isOK(checkRedPairUrAfProc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) [] [] P rP R (computeRuleMap R) ur ?af)" (is "isOK(?proc)") 
    by (auto simp: Let_def) 
  have "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(checkRedPairUrAfProc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) [] [] P rP R rm ur ?af))"
    by (rule ce_af_redpair_impl.check_redpair_ur_af_proc_sound, rule conditional_inter_ce_af_redpair_impl, rule check1)
  thus "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
    unfolding chain_identifying_onlyDP_proc_impl_def
    using accept by best
qed
end

context mono_poly_order_carrier
begin

lemma create_rule_removal_proc_comp_sound: 
  assumes check: "isOK(create_rule_removal_proc_comp carrier_components check_mono I R rR)"
  and SN: "SN (rstep (set rR))"
  shows "SN (rstep (set R))"
proof -
  let ?c = carrier_components
  let ?pi = "pi_of I"
  from check have check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def)
  from check have accept: "isOK(checkRuleRemovalProc (check_poly_monoI 0 geq check_mono I) (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) [] [] R rR)" (is "isOK(?proc)") 
    by (auto simp: Let_def) 
  show "SN (rstep (set R))"
    by (rule mono_redpair_impl.checkRuleRemovalProcSound, rule conditional_inter_mono_redpair_impl, rule check1, rule accept, rule SN)
qed

lemma mono_redpair_proc_comp_ci : 
  "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp carrier_components check_mono I P rP R rR))"
unfolding "chain_identifying_proc_impl_def"
proof (intro allI, intro impI)
  fix P R rP rR
  let ?c = carrier_components
  let ?pi = "pi_of I"
  assume check: "isOK(create_mono_redpair_proc_comp ?c check_mono I P rP R rR)"
  hence check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def)
  from check have accept: "isOK(checkMonoRedPairProc (check_poly_monoI 0 geq check_mono I) (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) [] [] P rP R rR)" (is "isOK(?proc)") 
    by (auto simp: Let_def) 
  have "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(checkMonoRedPairProc (check_poly_mono I) (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) [] [] P rP R rR))"
    by (rule mono_redpair_impl.check_mono_redpair_proc_sound, rule conditional_inter_mono_redpair_impl, rule check1)
  thus "set rR \<subseteq> set R \<and> (\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))"
    unfolding chain_identifying_proc_impl_def
    using accept by best
qed

lemma mono_redpair_ur_proc_comp_ci : 
  "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp carrier_components check_mono I P rP R rm rR ur))"
unfolding "chain_identifying_proc_impl_def"
proof (intro allI, intro impI)
  fix P R rP rR
  let ?c = carrier_components
  let ?pi = "pi_of I"
  assume check: "isOK(create_mono_redpair_ur_proc_comp ?c check_mono I P rP R (computeRuleMap R) rR ur)"
  hence check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def)
  from check have check2: "isOK(check_poly_mono I)" by (auto simp: Let_def)
  from check have accept: "isOK(checkMonoRedPairUrProc (check_poly_monoI 0 geq check_mono I) (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) [] [] P rP R (computeRuleMap R) rR ur)" (is "isOK(?proc)") 
    by (auto simp: Let_def) 
  have "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(checkMonoRedPairUrProc (check_poly_mono I) (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) [] [] P rP R rm rR ur))"
    by (rule ce_mono_redpair_impl.check_mono_redpair_ur_proc_sound,
      rule conditional_inter_ce_mono_redpair_impl, rule check1, rule check2)
  thus "set rR \<subseteq> set R \<and> (\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))"
    unfolding chain_identifying_proc_impl_def
    using accept by best
qed
end



end


