(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory ProofChecker imports SubtermCriterionImpl DepGraph RedPairImpl DpTransImpl Loop SemanticLabeling begin



text {* 
   A polynomial order is represented by a list of function symbol assignments where 
   every function symbol obtains a pair of a number n and a list of numbers [n1,..,nk].
   The corresponding polynomial interpretation is as follows:
   Pol(f(x1,..,xl)) = n + sum_i ni * xi
        where i ranges from 1 to min(k,l)
*}

types ('f,'l,'v)trsLL = "(('f,'l)lab,'v)trsL"

datatype ('f,'l,'v)"DPProof" = PisEmpty  
                          | RedPairProc "('f,'l)lab redpair" "('f,'l,'v)trsLL"  "('f,'l,'v)DPProof"
                          | RedPairUrProc "('f,'l)lab redpair" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v) DPProof"
                          | DepGraphProc "(('f,'l,'v)DPProof option \<times> ('f,'l,'v)trsLL) list"
                          | MonoRedPairProc "('f,'l)lab redpair" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" 
                          | MonoRedPairUrProc "('f,'l)lab redpair" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" 
                          | UnlabProc "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof"
                          | SemlabProc "('f,'l)sl_inter" "('f,nat,'l)label" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof"
                          | SpscProc "('f,'l)lab spL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof"


fun indent :: "sstring list \<Rightarrow> sstring list"
where "indent xs = (newline # (map (\<lambda> s. ''  '' @ s) xs))" 

fun debug :: "sstring \<Rightarrow> sstring \<Rightarrow> 'a \<Rightarrow> 'a"
where "debug i t x = x"

types ('f,'l,'v)ruleMap = "(('f,'l)lab \<Rightarrow> ('f,'l,'v)trsLL)option"

function checkDPProof :: "sstring \<Rightarrow> ('f,'l,'v)trsLL \<Rightarrow> ('f,'l,'v)trsLL \<Rightarrow> ('f,'l,'v)ruleMap \<Rightarrow> ('f,'l,'v)ruleMap \<Rightarrow> ('f :: sshow, 'l :: sshow, 'v :: sshow)DPProof \<Rightarrow> sstring list monad" 
where 
   "checkDPProof i P R rm rrm PisEmpty = (debug i ''P is empty''
            (case P of Nil \<Rightarrow> return () | Cons _ _ \<Rightarrow> 
                     error ((i +++ '': P is not empty in the following DP-problem '') # showDPP P R)))"  
|  "checkDPProof i P R rm rrm (RedPairProc redp rP prf)  = (debug i ''RedPairProc'' (
         create_redpair_proc redp P rP R 
           \<leftarrow> (\<lambda> s. ((i +++ '': error when applying the reduction pair processor to switch from '') # (showDPP P R @ ''to the DP problem '' # showDPP rP R @ s))) 
       \<guillemotright> checkDPProof (i +++ ''.1'') rP R rm rrm prf
           \<leftarrow> (\<lambda> s. (i +++ '': error below the reduction pair processor'') # (indent s))
))" 
|  "checkDPProof i P R rm rrm (RedPairUrProc redp rP ur prf)  = (debug i ''RedPairProc (usable rules + af)'' (
         create_redpair_ur_af_proc redp P rP R rm ur
           \<leftarrow> (\<lambda> s. ((i +++ '': error when applying the reduction pair processor (with usable rules and af) to switch from '') # (showDPP P R @ ''to the DP problem '' # showDPP rP R @ s))) 
       \<guillemotright> checkDPProof (i +++ ''.1'') rP R rm rrm prf
           \<leftarrow> (\<lambda> s. (i +++ '': error below the reduction pair processor'') # (indent s))
))" 
| "checkDPProof i P R rm rrm (DepGraphProc edpts) = (debug i ''DepGraphProc'' (
       checkDepGraphProc P (map (\<lambda> x. ((case fst x of Some _ \<Rightarrow> True | None \<Rightarrow> False), snd x)) edpts) rm rrm
           \<leftarrow> (\<lambda> s. (i +++ '': error while trying to perform SCC-decomposition  on '') # 
                          (showDPP P R @  s))
     \<guillemotright> checkAllMIndex (\<lambda> j rp. case fst rp of Some prf \<Rightarrow> Suc j | None \<Rightarrow> j) (\<lambda> j rp. case fst rp of Some prf \<Rightarrow> checkDPProof (i +++ ''.'' +++ sshow j) (snd rp) R rm rrm prf | None \<Rightarrow> OK) edpts
           \<leftarrow> (\<lambda> s. (i +++ '': error below the dependency graph processor'') # (indent s))
))"
|  "checkDPProof i P R rm rrm (MonoRedPairProc redp rP rR prf)  = (debug i ''MonoRedPairProc'' (
         create_mono_redpair_proc redp P rP R rR
           \<leftarrow> (\<lambda> s. ((i +++ '': error when applying the monotonic reduction pair processor to switch from '') # (showDPP P R @ ''to the DP problem '' # showDPP rP rR @ s))) 
       \<guillemotright> checkDPProof (i +++ ''.1'') rP rR (computeRuleMap rR) (computeRuleMap (reverse_trs rR)) prf
           \<leftarrow> (\<lambda> s. (i +++ '': error below the monotonic reduction pair processor'') # (indent s))
))"
|  "checkDPProof i P R rm rrm (MonoRedPairUrProc redp rP rR ur prf)  = (debug i ''MonoRedPairProc (usable rules)'' (
         create_mono_redpair_ur_proc redp P rP R rm rR ur
           \<leftarrow> (\<lambda> s. ((i +++ '': error when applying the monotonic reduction pair processor (with usable rules) to switch from '') # (showDPP P R @ ''to the DP problem '' # showDPP rP rR @ s))) 
       \<guillemotright> checkDPProof (i +++ ''.1'') rP rR (computeRuleMap rR) (computeRuleMap (reverse_trs rR)) prf
           \<leftarrow> (\<lambda> s. (i +++ '': error below the monotonic reduction pair (with usable rules) processor'') # (indent s))
))"
|  "checkDPProof i P R rm rrm (UnlabProc ulP ulR prf)  = (debug i ''UnlabProc'' (
        check_unlab_proc P R ulP ulR
      \<guillemotright> checkDPProof (i +++ ''.1'') ulP ulR (computeRuleMap ulR) (computeRuleMap (reverse_trs ulR)) prf
           \<leftarrow> (\<lambda> s. (i +++ '': error below the unlabeling processor'') # (indent s)) 
))"
|  "checkDPProof i P R rm rrm (SemlabProc sli L lP lR prf)  = (debug i ''SemlabProc'' (
       create_sem_lab_proc sli L P R lP lR
           \<leftarrow> (\<lambda> s. ((i +++ '': error when applying the semantic labeling processor'') # s))
    \<guillemotright>  checkDPProof (i +++ ''.1'') lP lR (computeRuleMap lR) (computeRuleMap (reverse_trs lR)) prf
           \<leftarrow> (\<lambda> s. (i +++ '': error below the semantic labeling processor'') # (indent s))
))"
|  "checkDPProof i P R rm rrm (SpscProc sp rP prf) = (debug i ''SpscProc'' (
       create_spsc_proc sp P rP R
           \<leftarrow> (\<lambda> s. ((i +++ '': error when applying the subterm criterion processor'') # s))
    \<guillemotright>  checkDPProof (i +++ ''.1'') rP R rm rrm prf
           \<leftarrow> (\<lambda> s. (i +++ '': error below the subterm criterion processor'') # (indent s))
))"
by (pat_completeness, auto)
termination "checkDPProof"
proof 
  let ?R = "measures [(\<lambda> (i,P,R,rm,rrm,prf). size prf)]"
  show "wf ?R" by auto
next
  fix i P R rm rrm redp rP "prf" 
  show "((i +++ ''.1'', rP, R, rm, rrm, prf), i, P, R, rm, rrm, RedPairProc redp rP prf) \<in> measures [\<lambda>(i, P, R, rm, rrm, prf). size prf]" by auto
next
  fix i P R rm rrm redp rP ur "prf"
  show "((i +++ ''.1'', rP, R, rm, rrm, prf), i, P, R, rm, rrm, RedPairUrProc redp rP ur prf) \<in> measures [\<lambda>(i, P, R, rm, rrm, prf). size prf]" by auto
next
  fix i and P R :: "('f::sshow, 'l::sshow,'v:: sshow)trsLL" and rm rrm and  
    edpts :: "(('f,'l,'v)DPProof option \<times> ('f,'l,'v)trsLL) list" and n x and "prf" :: "('f,'l,'v)DPProof"
  assume elem: "x \<in> set edpts" and "fst x = Some prf"
  hence "(option_size size) (fst x) = Suc (size prf)" by auto
  hence "prod_size (option_size size) (list_size
                (prod_size (Term.term.term_size (lab_size (\<lambda>x. 0) (\<lambda>x. 0)) (\<lambda>x. 0)) 
                (Term.term.term_size (lab_size (\<lambda>x. 0) (\<lambda>x. 0)) (\<lambda>x. 0)))) x > size prf" 
    (is "?m x > size prf")
    by (cases x, auto)   
  with elem have "size prf < ((list_size ?m) edpts)" 
  proof (induct edpts, simp)
    case (Cons y ys)
    thus ?case by (cases "x = y", auto)
  qed
  with elem show "((i +++ ''.'' +++ sshow n, snd x, R, rm, rrm, prf), i, P, R, rm, rrm, DepGraphProc edpts) \<in> measures [\<lambda>(i, P, R, rm, rrm, prf). size prf]"  by auto
next
  fix i P R rm rrm redp rP rR "prf"
  show "((i +++ ''.1'', rP, rR, computeRuleMap rR,
         computeRuleMap (reverse_trs rR), prf),
        i, P, R, rm, rrm, MonoRedPairProc redp rP rR prf)
       \<in> measures [\<lambda>(i, P, R, rm, rrm, prf). size prf]" by auto
next
  fix i P R rm rrm redp rP rR ur "prf"
  show "((i +++ ''.1'', rP, rR, computeRuleMap rR,
         computeRuleMap (reverse_trs rR), prf),
        i, P, R, rm, rrm, MonoRedPairUrProc redp rP rR ur prf)
       \<in> measures [\<lambda>(i, P, R, rm, rrm, prf). size prf]" by auto
next
  fix i P R rm rrm ulP ulR "prf" 
  show " ((i +++ ''.1'', ulP, ulR, computeRuleMap ulR,
         computeRuleMap (reverse_trs ulR), prf),
        i, P, R, rm, rrm, UnlabProc ulP ulR prf)
       \<in> measures [\<lambda>(i, P, R, rm, rrm, prf). size prf]" by auto
next
  fix i P R rm rrm sli L lP lR "prf"
  show "((i +++ ''.1'', lP, lR, computeRuleMap lR,
         computeRuleMap (reverse_trs lR), prf),
        i, P, R, rm, rrm, SemlabProc sli L lP lR prf)
       \<in> measures [\<lambda>(i, P, R, rm, rrm, prf). size prf]" by auto
next
  fix i P R rm rrm sp rP "prf"
  show "((i +++ ''.1'',rP,R,rm,rrm,prf),i,P,R,rm,rrm,
    SpscProc sp rP prf)
    \<in> measures [\<lambda>(i,P,R,rm,rrm,prf). size prf]" by auto
qed


lemma checkDPProofSoundI  : fixes P :: "('f :: sshow, 'l :: sshow,'v :: sshow)trsLL"
  shows "isOK(checkDPProof i P R rm rrm prf) \<and>
    rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R) \<longrightarrow> finite_lab_dpp (set P, set R)"
proof (induct rule:checkDPProof.induct[where P = "\<lambda> i P R rm rrm prf. isOK(checkDPProof i P R rm rrm prf) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R) \<longrightarrow> finite_lab_dpp (set P, set R)"])
  case (1 i P R rm rrm)  thus ?case unfolding finite_lab_dpp_def by (cases P, auto)
next
  case (2 i P R rm rrm redp rP prof) 
  {
    assume rm: "isOK (checkDPProof i P R rm rrm (RedPairProc redp rP prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R)"
    proof (rule apply_ci_onlyDP_lab_proc[where rP = rP],  rule create_redpair_proc) 
      from rm 2 show "isOK (create_redpair_proc redp P rP R)" by auto
    next
      from rm 2 show "finite_lab_dpp (set rP, set R)" by force
    qed
  }
  thus ?case by blast
next
  case (3 i P R rm rrm redp rP ur prof)
  {
    assume rm: "isOK (checkDPProof i P R rm rrm (RedPairUrProc redp rP ur prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R)"
    proof (rule apply_ci_onlyDP_lab_proc[where rP = rP], rule create_redpair_ur_af_proc)
      from rm 3 show "isOK (create_redpair_ur_af_proc redp P rP R (computeRuleMap R) ur)" by auto
    next
      from rm 3 show "finite_lab_dpp (set rP, set R)" by force
    qed
  }
  thus ?case by blast
next
  case (4 i P R rm rrm edpts) 
  let ?mc = "\<lambda> edpts. (map (\<lambda> x. ((case fst x of Some _ \<Rightarrow> True | None \<Rightarrow> False), snd x)) edpts)"
  let ?edps = "?mc edpts"
  {
    assume rm: "isOK (checkDPProof i P R rm rrm (DepGraphProc edpts)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R)"
    proof (rule dep_graph_lab_proc, auto)
      from 4 rm show "isOK(checkDepGraphProc P ?edps (computeRuleMap R) (computeRuleMap (reverse_trs R)))" by (auto)
    next
      fix a rp
      assume elem: "(True, rp) \<in> set ?edps"
      from 4 rm have "isOK ( checkAllMIndex (\<lambda> j rp. case fst rp of Some prf \<Rightarrow> Suc j | None \<Rightarrow> j) (\<lambda> j rp. case fst rp of Some prf \<Rightarrow> checkDPProof (i +++ ''.'' +++ sshow j) (snd rp) R rm rrm prf | None \<Rightarrow> OK) edpts)" (is "isOK ( checkAllMIndex ?g ?f edpts)") by auto
      hence "\<forall> prfrP \<in> set edpts. \<exists> n. isOK ( ?f n prfrP)" by (rule checkAllMIndexSound)
      with 4 rm have fin: "\<forall> prf rP. (Some prf, rP) \<in> set edpts \<longrightarrow> finite_lab_dpp (set rP, set R)" by auto
      from elem obtain sub where "set sub \<subseteq> set edpts \<and> (True, rp) \<in> set (?mc sub)" by blast
      thus "finite_lab_dpp (set rp, set R)"
      proof (induct sub, simp)
	case (Cons edp subb)
	with fin show ?case 
	  by (cases "edp", cases "fst edp", simp, cases "snd edp = rp", auto)
      qed
    qed
  }
  thus ?case by blast
next
  case (5 i P R rm rrm redp rP rR prof)
  {
    assume rm: "isOK (checkDPProof i P R rm rrm (MonoRedPairProc redp rP rR prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R)"
    proof (rule apply_ci_lab_proc[where rP = rP], rule create_mono_redpair_proc)
      from rm 5 show "isOK (create_mono_redpair_proc redp P rP R rR)" by auto
    next
      from rm 5 show "finite_lab_dpp (set rP, set rR)" by force
    qed
  }
  thus ?case by blast
next
  case (6 i P R rm rrm redp rP rR ur prof)
  {
    assume rm: "isOK (checkDPProof i P R rm rrm (MonoRedPairUrProc redp rP rR ur prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R)"
    proof (rule apply_ci_lab_proc[where rP = rP], rule create_mono_redpair_ur_proc)
      from rm 6 show "isOK (create_mono_redpair_ur_proc redp P rP R (computeRuleMap R) rR ur)" by auto
    next
      from rm 6 show "finite_lab_dpp (set rP, set rR)" by force
    qed
  }
  thus ?case by blast
next
  case (7 i P R rm rrm ulP ulR prof)
  {
    assume rm: "isOK(checkDPProof i P R rm rrm (UnlabProc ulP ulR prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R)"
    proof (rule apply_sound_lab_proc[where rP = ulP and rR = ulR], rule check_unlab_proc_sound)
      from rm 7 show "finite_lab_dpp (set ulP, set ulR)" by (force simp: Let_def)
    next
      from rm 7 show "isOK(check_unlab_proc P R ulP ulR)" by auto
    qed
  }
  thus ?case by blast
next
  case (8 i P R rm rrm sli L lP lR prof)
  {
    assume rm: "isOK (checkDPProof i P R rm rrm (SemlabProc sli L lP lR prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R)" 
    proof (rule apply_sound_lab_proc[where rP = lP], rule create_sem_lab_proc_sound)
      from rm 8 show "isOK (create_sem_lab_proc sli L P R lP lR)" by auto
    next
      from rm 8 show "finite_lab_dpp (set lP, set lR)" by force
    qed
  }
  thus ?case by blast
next
  case (9 i P R rm rrm sp rP prof)
  {
    assume rm: "isOK(checkDPProof i P R rm rrm (SpscProc sp rP prof))
      \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp(set P,set R)"
    proof (rule apply_ci_onlyDP_lab_proc[where rP = rP], rule spsc_proc_ci)
      from rm 9 show "isOK(create_spsc_proc sp P rP R)" by auto
    next
      from rm 9 show "finite_lab_dpp(set rP,set R)" by force
    qed
  }
  thus ?case by blast
qed

declare checkDPProof.simps[simp del]

lemma checkDPProofSound: "isOK(checkDPProof i P R (computeRuleMap R) (computeRuleMap (reverse_trs R)) prf) \<Longrightarrow> finite_lab_dpp (set P, set R)"
  using checkDPProofSoundI by blast

datatype ('f,'l,'v)"TRSProof" = DPTrans "(('f,'l)lab shp,'v)trsL"  "(('f,'l)lab shp,'l,'v) DPProof" 
                              | RuleRemoval "('f,'l)lab redpair" "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof"
                              | Unlab "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof"
                              | Semlab "('f,'l)sl_inter" "('f,nat,'l)label" "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof"
                              | RisEmpty


fun checkTRSProof :: "sstring \<Rightarrow> ('f :: sshow, 'l :: sshow, 'v :: sshow)trsLL \<Rightarrow> ('f,'l,'v) TRSProof \<Rightarrow> sstring list monad"
where "checkTRSProof i r (DPTrans P prf) =  (debug i ''DP trans'' (
    (\<lambda> IR. 
    ((\<lambda> ILR. checkDPs r P
      \<leftarrow> (\<lambda> s. (i +++ '': error when switching from the TRS'') # 
                 ( showTRS r @  (''to the initial DP problem'' # showDPP P IR @  s)))
  \<guillemotright> checkDPProof (i +++ ''.1'') (init_lab_trsL P) ILR (computeRuleMap ILR) (computeRuleMap (reverse_trs ILR)) prf
      \<leftarrow> (\<lambda> s. (i +++ '': error below switch to dependency pairs'') # (indent s))) (init_lab_trsL IR))) (computeID r)
))"
  | "checkTRSProof i r (RuleRemoval rp rR prf) = (debug i ''rule removal'' (
      create_rule_removal_proc rp r rR
           \<leftarrow> (\<lambda> s. ((i +++ '': error in rule removal when switching from the TRS'') # (showTRS r @ (''to the TRS'' # (showTRS rR @ s))))) 
       \<guillemotright> checkTRSProof (i +++ ''.1'') rR prf
           \<leftarrow> (\<lambda> s. (i +++ '': error below the rule removal processor'') # (indent s))
))"
  | "checkTRSProof i r (Unlab uR prf) = (debug i ''unlabel'' (
      check_unlab_trs r uR
           \<leftarrow> (\<lambda> s. ((i +++ '': error in unlabeling of TRS when switching from the TRS'') # (showTRS r @ (''to the TRS'' # (showTRS uR @ s))))) 
       \<guillemotright> checkTRSProof (i +++ ''.1'') uR prf
           \<leftarrow> (\<lambda> s. (i +++ '': error below the unlabeling'') # (indent s))
))"
  | "checkTRSProof i r RisEmpty = debug i ''R is empty'' (checkB (r = []) ([i +++ '': R is not empty, but it is the TRS''] @ showTRS r))"
  |  "checkTRSProof i R (Semlab sli L lR prf)  = (debug i ''Semlab'' (
       create_sem_lab_trs sli L R lR
           \<leftarrow> (\<lambda> s. ((i +++ '': error when applying semantic labeling'') # s))
    \<guillemotright>  checkTRSProof (i +++ ''.1'') lR prf
           \<leftarrow> (\<lambda> s. (i +++ '': error below semantic labeling'') # (indent s))
))"

lemma checkTRSProofSound  : fixes prof :: "('f :: sshow,'l :: sshow,'v :: sshow)TRSProof"
  assumes check: "isOK(checkTRSProof i r prof)"
  shows "SN (rstep (set r))"
using assms
proof (induct "prof" arbitrary:r i)
  case (DPTrans P dpprf)
  let ?il = "init_lab_trsL :: (('f,'l)lab shp,'v)trsL \<Rightarrow> (('f,'l)lab shp,'l,'v)trsLL"
  let ?ilr = "?il (computeID r)"
  from DPTrans have dpsSound: "isOK(checkDPs r P)" and rec: "isOK(checkDPProof (i +++ ''.1'') (?il P) ?ilr (computeRuleMap ?ilr) (computeRuleMap (reverse_trs ?ilr)) dpprf)" 
    by (auto simp: Let_def)
  from rec have "finite_lab_dpp (set (?il P), set (?il (computeID r)))" 
    by (rule checkDPProofSound) 
  with init_lab_sound have "finite_dpp (set P, set (computeID r))" by auto
  with dpsSound show "SN (rstep (set r))" by (simp)
next
  case (RuleRemoval rp rR prof')
  hence ok:"isOK (checkTRSProof i r (RuleRemoval rp rR prof'))" and IH:"\<And>r' i'. isOK (checkTRSProof i' r' prof') \<Longrightarrow> SN (rstep (set r'))" by auto
  from ok have rr_ok:"isOK(create_rule_removal_proc rp r rR)" and rec_ok:"isOK(checkTRSProof (i +++ ''.1'') rR prof')" by auto
  from rec_ok IH have "SN (rstep (set rR))" by auto
  with rr_ok show "SN (rstep (set r))" using create_rule_removal_proc by auto
next
  case RisEmpty
  hence "r = []" by auto
  hence "\<forall> l rr p \<sigma>. rstep_r_p_s (set r) (l,rr) p \<sigma> = {}" unfolding rstep_r_p_s_def by auto
  with rstep_iff_rstep_r_p_s have "rstep (set r) = {}" by best 
  thus "SN (rstep (set r))" unfolding SN_ars_def SN_elt_def by blast
next
  case (Unlab uR prof)
  hence ok:"isOK (checkTRSProof i r (Unlab uR prof))" and IH:"\<And>r' i'. isOK (checkTRSProof i' r' prof) \<Longrightarrow> SN (rstep (set r'))" by auto
  from ok have ul_ok:"isOK(check_unlab_trs r uR)" and rec_ok:"isOK(checkTRSProof (i +++ ''.1'') uR prof)" by auto
  from rec_ok IH have "SN (rstep (set uR))" by auto
  with ul_ok show "SN (rstep (set r))" using check_unlab_trs_sound by auto
next
  case (Semlab sli L lR prof)
  hence ok:"isOK (checkTRSProof i r (Semlab sli L lR prof))" and IH:"\<And>r' i'. isOK (checkTRSProof i' r' prof) \<Longrightarrow> SN (rstep (set r'))" by auto
  from ok have l_ok:"isOK(create_sem_lab_trs sli L r lR)" and rec_ok:"isOK(checkTRSProof (i +++ ''.1'') lR prof)" by auto
  from rec_ok IH have "SN (rstep (set lR))" by auto
  with l_ok show "SN (rstep (set r))" using create_sem_lab_trs_sound by auto
qed

datatype ('f,'v)"TRSDisproof" = Loop "('f,'v)loopL" | NotWellFormed


fun checkTRSDisproof :: "string \<Rightarrow> ('f :: sshow, 'v :: sshow)trsL \<Rightarrow> ('f,'v) TRSDisproof \<Rightarrow> sstring list monad"
where "checkTRSDisproof i r (Loop l) = debug i ''Loop'' (checkLoop r l \<leftarrow> (\<lambda> s. (i @ '': error in loop proof'') # s))"
    | "checkTRSDisproof i r NotWellFormed = debug i ''not well-formed'' (checkNotWellFormed r \<leftarrow> (\<lambda> s. (i @ '': error in not-well-formed proof'') # s))"

lemma checkTRSDisproofSound :
 assumes "isOK(checkTRSDisproof i r prf)"
 shows "\<not> SN (rstep (set r))"
using assms 
proof (cases "prf")
  case NotWellFormed
  with assms show ?thesis by simp
next
  case (Loop x)
  show ?thesis 
  proof (cases x)
    fix ts C \<sigma> 
    assume "x = (ts,C,\<sigma>)"
    with assms Loop have "isOK(checkLoop r (ts, C, \<sigma>))" by simp
    thus ?thesis
      by (cases "subst_safe [] \<sigma> \<iota>") auto 
  qed
qed

end
